import pytest

import geomstats.backend as gs
from geomstats.geometry.diffeo import AutodiffDiffeo
from geomstats.test.random import RandomDataGenerator
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data


class CircleSO2Diffeo(AutodiffDiffeo):
    """Diffeomorphism between S1 and SO2."""

    def __init__(self):
        super().__init__(space_shape=(2,), image_space_shape=(2, 2))

    def __call__(self, base_point):
        second_column = gs.stack([-base_point[..., 1], base_point[..., 0]], axis=-1)
        return gs.stack([base_point, second_column], axis=-1)

    def inverse(self, image_point):
        return image_point[..., 0]


class DiffeoTestCase(TestCase):
    tangent_to_multiple = False

    def setup_method(self):
        if not hasattr(self, "data_generator") and hasattr(self, "space"):
            self.data_generator = RandomDataGenerator(self.space)

        if not hasattr(self, "image_data_generator") and hasattr(self, "image_space"):
            self.image_data_generator = RandomDataGenerator(self.image_space)

    def test_diffeomorphism(self, base_point, expected, atol):
        res = self.diffeo(base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_diffeomorphism_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()

        expected = self.diffeo(base_point)

        vec_data = generate_vectorization_data(
            data=[dict(base_point=base_point, expected=expected, atol=atol)],
            arg_names=["base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_diffeomorphism_belongs(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        image_point = self.diffeo(base_point)

        belongs = self.image_space.belongs(image_point, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(belongs, expected)

    def test_inverse(self, image_point, expected, atol):
        res = self.diffeo.inverse(image_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_inverse_vec(self, n_reps, atol):
        image_point = self.image_data_generator.random_point()

        expected = self.diffeo.inverse(image_point)

        vec_data = generate_vectorization_data(
            data=[dict(image_point=image_point, expected=expected, atol=atol)],
            arg_names=["image_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_inverse_belongs(self, n_points, atol):
        image_point = self.image_data_generator.random_point(n_points)

        point = self.diffeo.inverse(image_point)

        belongs = self.space.belongs(point, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(belongs, expected)

    @pytest.mark.random
    def test_inverse_after_diffeomorphism(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        image_point = self.diffeo(base_point)
        base_point_ = self.diffeo.inverse(image_point)

        self.assertAllClose(base_point_, base_point, atol=atol)

    @pytest.mark.random
    def test_diffeomorphism_after_inverse(self, n_points, atol):
        image_point = self.image_data_generator.random_point(n_points)

        point = self.diffeo.inverse(image_point)
        image_point_ = self.diffeo(point)

        self.assertAllClose(image_point_, image_point, atol=atol)

    def test_tangent(
        self, tangent_vec, expected, atol, base_point=None, image_point=None
    ):
        res = self.diffeo.tangent(
            tangent_vec, base_point=base_point, image_point=image_point
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_tangent_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.diffeo.tangent(tangent_vec, base_point)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_tangent_is_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        image_tangent_vec = self.diffeo.tangent(tangent_vec, base_point)
        image_point = self.diffeo(base_point)

        is_tangent = self.image_space.is_tangent(
            image_tangent_vec, image_point, atol=atol
        )
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(is_tangent, expected)

    @pytest.mark.random
    def test_tangent_with_image_point(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        image_point = self.diffeo(base_point)

        image_tangent_vec = self.diffeo.tangent(tangent_vec, base_point)
        image_tangent_vec_ = self.diffeo.tangent(tangent_vec, image_point=image_point)

        self.assertAllClose(image_tangent_vec, image_tangent_vec_, atol=atol)

    def test_inverse_tangent(
        self,
        image_tangent_vec,
        expected,
        atol,
        image_point=None,
        base_point=None,
    ):
        res = self.diffeo.inverse_tangent(
            image_tangent_vec,
            image_point=image_point,
            base_point=base_point,
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_inverse_tangent_vec(self, n_reps, atol):
        image_point = self.image_data_generator.random_point()
        image_tangent_vec = self.image_data_generator.random_tangent_vec(image_point)

        expected = self.diffeo.inverse_tangent(image_tangent_vec, image_point)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    image_tangent_vec=image_tangent_vec,
                    image_point=image_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["image_tangent_vec", "image_point"],
            expected_name="expected",
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_inverse_tangent_is_tangent(self, n_points, atol):
        image_point = self.image_data_generator.random_point(n_points)
        image_tangent_vec = self.image_data_generator.random_tangent_vec(image_point)

        tangent_vec = self.diffeo.inverse_tangent(image_tangent_vec, image_point)
        base_point = self.diffeo.inverse(image_point)

        is_tangent = self.space.is_tangent(tangent_vec, base_point, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(is_tangent, expected)

    @pytest.mark.random
    def test_inverse_tangent_with_base_point(self, n_points, atol):
        image_point = self.image_data_generator.random_point(n_points)
        image_tangent_vec = self.image_data_generator.random_tangent_vec(image_point)

        base_point = self.diffeo.inverse(image_point)

        tangent_vec = self.diffeo.inverse_tangent(image_tangent_vec, image_point)
        tangent_vec_ = self.diffeo.inverse_tangent(
            image_tangent_vec, base_point=base_point
        )
        self.assertAllClose(tangent_vec, tangent_vec_, atol=atol)

    @pytest.mark.random
    def test_inverse_tangent_after_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        image_tangent_vec = self.diffeo.tangent(tangent_vec, base_point)
        image_point = self.diffeo(base_point)

        tangent_vec_ = self.diffeo.inverse_tangent(image_tangent_vec, image_point)
        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)

    @pytest.mark.random
    def test_tangent_after_inverse_tangent(self, n_points, atol):
        image_point = self.image_data_generator.random_point(n_points)
        image_tangent_vec = self.image_data_generator.random_tangent_vec(image_point)

        tangent_vec = self.diffeo.inverse_tangent(image_tangent_vec, image_point)
        base_point = self.diffeo.inverse(image_point)

        image_tangent_vec_ = self.diffeo.tangent(tangent_vec, base_point)
        self.assertAllClose(image_tangent_vec_, image_tangent_vec, atol=atol)


class AutodiffDiffeoTestCase(DiffeoTestCase):
    def test_jacobian(self, base_point, expected, atol):
        res = self.diffeo.jacobian(base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_jacobian_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()

        expected = self.diffeo.jacobian(base_point)

        vec_data = generate_vectorization_data(
            data=[dict(base_point=base_point, expected=expected, atol=atol)],
            arg_names=["base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_inverse_jacobian(self, image_point, expected, atol):
        res = self.diffeo.inverse_jacobian(image_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_inverse_jacobian_vec(self, n_reps, atol):
        image_point = self.image_data_generator.random_point()

        expected = self.diffeo.inverse_jacobian(image_point)

        vec_data = generate_vectorization_data(
            data=[dict(image_point=image_point, expected=expected, atol=atol)],
            arg_names=["image_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)


class DiffeoComparisonTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator") and hasattr(self, "space"):
            self.data_generator = RandomDataGenerator(self.space)

        if not hasattr(self, "image_data_generator") and hasattr(self, "image_space"):
            self.image_data_generator = RandomDataGenerator(self.image_space)

    @pytest.mark.random
    def test_diffeomorphism(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        res = self.diffeo(base_point)
        res_ = self.other_diffeo(base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_inverse(self, n_points, atol):
        image_point = self.image_data_generator.random_point(n_points)

        res = self.diffeo.inverse(image_point)
        res_ = self.other_diffeo.inverse(image_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        res = self.diffeo.tangent(tangent_vec, base_point)
        res_ = self.other_diffeo.tangent(tangent_vec, base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_inverse_tangent(self, n_points, atol):
        image_point = self.image_data_generator.random_point(n_points)
        image_tangent_vec = self.image_data_generator.random_tangent_vec(image_point)

        res = self.diffeo.inverse_tangent(image_tangent_vec, image_point)
        res_ = self.other_diffeo.inverse_tangent(image_tangent_vec, image_point)
        self.assertAllClose(res, res_, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.base import VectorSpaceOpenSetTestCase
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase


class PoincareBallTestCase(VectorSpaceOpenSetTestCase):
    def test_projection_norm_less_than_1(self, point, atol):
        projected_point = self.space.projection(point)
        result = gs.sum(projected_point * projected_point, axis=-1) < 1.0 + atol
        self.assertTrue(result)


class PoincareBallMetricTestCase(RiemannianMetricTestCase):
    # TODO: complete tests
    def test_mobius_add(self, point_a, point_b, project_first, expected, atol):
        res = self.space.metric.mobius_add(
            point_a, point_b, project_first=project_first
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_mobius_add_vec(self, n_reps, atol):
        # TODO: check project first false
        project_first = True
        point_a = self.data_generator.random_point()
        point_b = self.data_generator.random_point()

        expected = self.space.metric.mobius_add(
            point_a,
            point_b,
            project_first=project_first,
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    point_a=point_a,
                    point_b=point_b,
                    project_first=project_first,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["point_a", "point_b"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_retraction(self, tangent_vec, base_point, expected, atol):
        res = self.space.metric.retraction(tangent_vec, base_point)
        self.assertAllClose(res, expected, atol=atol)



"""The vector space of symmetric matrices.

Lead author: Yann Thanwerdas.
"""

import geomstats.backend as gs
from geomstats.geometry.base import (
    DiffeomorphicMatrixVectorSpace,
    LevelSet,
    MatrixVectorSpace,
)
from geomstats.geometry.diffeo import Diffeo
from geomstats.geometry.euclidean import EuclideanMetric
from geomstats.geometry.matrices import Matrices, MatricesMetric
from geomstats.vectorization import repeat_out, repeat_out_multiple_ndim


class SymmetricMatrices(MatrixVectorSpace):
    """Class for the vector space of symmetric matrices of size n.

    Parameters
    ----------
    n : int
        Integer representing the shapes of the matrices: n x n.
    """

    def __init__(self, n, equip=True):
        self.n = n
        super().__init__(dim=int(n * (n + 1) / 2), shape=(n, n), equip=equip)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return MatricesMetric

    def _create_basis(self):
        """Compute the basis of the vector space of symmetric matrices."""
        indices, values = [], []
        k = -1
        for row in range(self.n):
            for col in range(row, self.n):
                k += 1
                if row == col:
                    indices.append((k, row, row))
                    values.append(1.0)
                else:
                    indices.extend([(k, row, col), (k, col, row)])
                    values.extend([1.0, 1.0])

        return gs.array_from_sparse(indices, values, (k + 1, self.n, self.n))

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a matrix is symmetric.

        Parameters
        ----------
        point : array-like, shape=[.., n, n]
            Point to test.
        atol : float
            Tolerance to evaluate equality with the transpose.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the space.
        """
        belongs = super().belongs(point)
        if gs.any(belongs):
            is_symmetric = Matrices.is_symmetric(point, atol)
            return gs.logical_and(belongs, is_symmetric)
        return belongs

    def projection(self, point):
        """Make a matrix symmetric, by averaging with its transpose.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        sym : array-like, shape=[..., n, n]
            Symmetric matrix.
        """
        return Matrices.to_symmetric(point)

    @staticmethod
    def basis_representation(matrix_representation):
        """Convert a symmetric matrix into a vector.

        Parameters
        ----------
        matrix_representation : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        basis_representation : array-like, shape=[..., n(n+1)/2]
            Vector.
        """
        return gs.triu_to_vec(matrix_representation)

    @staticmethod
    def matrix_representation(basis_representation):
        """Convert a vector into a symmetric matrix.

        Parameters
        ----------
        basis_representation : array-like, shape=[..., n(n+1)/2]
            Vector.

        Returns
        -------
        matrix_representation : array-like, shape=[..., n, n]
            Symmetric matrix.
        """
        vec_dim = basis_representation.shape[-1]
        mat_dim = (gs.sqrt(8.0 * vec_dim + 1) - 1) / 2
        if mat_dim != int(mat_dim):
            raise ValueError(
                "Invalid input dimension, it must be of the form "
                "(n_samples, n * (n + 1) / 2)"
            )
        mat_dim = int(mat_dim)
        shape = (mat_dim, mat_dim)
        mask = 2 * gs.ones(shape) - gs.eye(mat_dim)
        indices = list(zip(*gs.triu_indices(mat_dim)))
        if gs.ndim(basis_representation) == 1:
            upper_triangular = gs.array_from_sparse(
                indices, basis_representation, shape
            )
        else:
            upper_triangular = gs.stack(
                [
                    gs.array_from_sparse(indices, data, shape)
                    for data in basis_representation
                ]
            )

        mat = Matrices.to_symmetric(upper_triangular) * mask
        return mat


class SymmetricHollowMatrices(LevelSet, MatrixVectorSpace):
    r"""Space of symmetric hollow matrices.

    Set of symmetric matrices with null diagonal:

    .. math::

        \operatorname{Hol}(n) = \{X \in \operatorname{Sym}(n)
        \mid \operatorname{Diag}(X)=0\}

    Parameters
    ----------
    n : int
        Integer representing the shapes of the matrices: n x n.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __init__(self, n, equip=True):
        self.n = n
        super().__init__(dim=int(n * (n - 1) / 2), shape=(n, n), equip=equip)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return MatricesMetric

    def _define_embedding_space(self):
        return SymmetricMatrices(n=self.n)

    def submersion(self, point):
        """Submersion that defines the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]

        Returns
        -------
        submersed_point : array-like, shape=[..., n]
        """
        return Matrices.diagonal(point)

    def tangent_submersion(self, vector, point):
        """Tangent submersion.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
        point : Ignored.

        Returns
        -------
        submersed_vector : array-like, shape=[..., n]
        """
        out = self.submersion(vector)
        return repeat_out(self.point_ndim, out, vector, point, out_shape=(self.n,))

    def _create_basis(self):
        """Compute the basis of the vector space of hollow symmetric matrices."""
        indices, values = [], []
        k = -1
        for row in range(self.n):
            for col in range(row + 1, self.n):
                k += 1
                indices.extend([(k, row, col), (k, col, row)])
                values.extend([1.0, 1.0])

        return gs.array_from_sparse(indices, values, (k + 1, self.n, self.n))

    @staticmethod
    def basis_representation(matrix_representation):
        """Convert a hollow symmetric matrix into a vector.

        Parameters
        ----------
        matrix_representation : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        vec : array-like, shape=[..., n(n+1)/2]
            Vector.
        """
        return gs.triu_to_vec(matrix_representation, k=1)

    def projection(self, point):
        """Project a point in embedding manifold on embedded manifold.

        Parameters
        ----------
        point : array-like, shape=[..., *embedding_space.point_shape]
            Point in embedding manifold.

        Returns
        -------
        projected : array-like, shape=[..., *point_shape]
            Projected point.
        """
        return point - Matrices.to_diagonal(point)


class HollowMatricesPermutationInvariantMetric(EuclideanMetric):
    r"""A permutation-invariant metric on the space of hollow matrices.

    It is flat Riemannian metric invariant by the congruence action
    of permutation matrices defined over a matrix vector space.

    Its associated quadratic form is:

    .. math::

        q(X)=\alpha \operatorname{tr}\left(X^2\right)
        +\beta \operatorname{Sum}\left(X^2\right)
        +\gamma \operatorname{Sum}(X)^2

    Parameters
    ----------
    space : Manifold
    alpha : float
        Scalar multiplying first term of quadratic form.
    beta : float
        Scalar multiplying second term of quadratic form.
    gamma : float
        Scalar multiplying third term of quadratic form.

    Check out chapter 8 of [T2022]_ for more details.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    """

    def __init__(self, space, alpha=None, beta=None, gamma=1.0):
        if alpha is None:
            alpha = 1.0 if space.n > 3 else 0.0

        if beta is None:
            beta = 1.0 if space.n > 2 else 0.0

        self._check_params(space, alpha, beta, gamma)
        super().__init__(space=space)
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

    @staticmethod
    def _check_params(space, alpha, beta, gamma):
        r"""Check parameters of quadratic form.

        The following conditions must verify:
        - n > 3: :math:`\alpha>0,2 \alpha+(n-2) \beta>0, \alpha+(n-1)(\beta+n \gamma)>0`
        - n = 3: :math:`\alpha=0, \beta > 0, \beta+3 \gamma>0`
        - n = 2: :math:`\alpha=0, \beta=0, \gamma > 0`
        """
        n = space.n
        if n == 2:
            if alpha > gs.atol or beta > gs.atol or gamma < gs.atol:
                raise ValueError(
                    f"When n==2: alpha ({alpha}) and beta ({beta}) must be 0, "
                    f"and gamma ({gamma}) > 0. "
                    "Check thanwerdas2022 theorem 8.7"
                )
            return

        elif n == 3:
            cond = beta + 3 * gamma
            if alpha > gs.atol or beta < gs.atol or cond < gs.atol:
                raise ValueError(
                    f"When n==3: alpha ({alpha}) must be 0, beta ({beta}) > 0 "
                    f"and an inequality greater than 0: {cond}. "
                    "Check thanwerdas2022 theorem 8.7"
                )
            return

        cond_1 = 2 * alpha + (n - 2) * beta
        cond_2 = alpha + (n - 1) * (beta + n * gamma)
        if cond_1 < gs.atol or cond_2 < gs.atol:
            raise ValueError(
                f"Inequalities should be greater than 0, but: {cond_1} and {cond_2}. "
                "Check thanwerdas2022 theorem 8.7"
            )

    def _quadratic_form(self, tangent_vec):
        """Quadratic form associated to inner product.

        Parameters
        ----------
        tangent_vec: array-like, shape=[..., n, n]
            Tangent vector at base point.
        """
        comp = gs.matmul(tangent_vec, tangent_vec)
        out_alpha = self.alpha * gs.trace(comp) if self.alpha > gs.atol else 0.0
        out_beta = (
            self.beta * gs.sum(comp, axis=(-2, -1)) if self.beta > gs.atol else 0.0
        )
        out_gamma = self.gamma * gs.sum(tangent_vec, axis=(-2, -1)) ** 2

        return out_alpha + out_beta + out_gamma

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a: array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b: array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point: array-like, shape=[..., n, n]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        inner_prod = (1 / 2) * (
            self._quadratic_form(tangent_vec_a + tangent_vec_b)
            - self._quadratic_form(tangent_vec_a)
            - self._quadratic_form(tangent_vec_b)
        )
        return repeat_out(
            self._space.point_ndim, inner_prod, tangent_vec_a, tangent_vec_b, base_point
        )

    def squared_norm(self, vector, base_point=None):
        """Compute the square of the norm of a vector.

        Squared norm of a vector associated to the inner product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
            Vector.
        base_point : array-like, shape=[..., n, n]
            Base point.
            Optional, default: None.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm.
        """
        return self._quadratic_form(vector)


class ConstantValueRowSumsDiffeo(Diffeo):
    r"""A diffeomorphism from the constant-value-row-sum matrices to symmetric matrices.

    A particular case is the diffeomorphism between the space of null-row-sum symmetric
    n-matrices and the space of symmetric (n-1)-matrices.

    Let :math:`f` be the diffeomorphism
    :math:`f: M \rightarrow N` of the manifold
    :math:`M` into the manifold :math:`N`.
    """

    def __init__(self, value=0.0):
        self.value = value
        self._space_ndim = 2
        self._image_space_ndim = 2

    def __call__(self, base_point):
        """Diffeomorphism at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        image_point : array-like, shape=[..., n-1, n-1]
            Image point.
        """
        return base_point[..., :-1, :-1]

    def _concatenate_row_sums(self, image_point, value):
        """Concatenate missing row and column."""
        row_sums = value - gs.sum(image_point, axis=-1)
        last_row = gs.concatenate(
            [row_sums, gs.expand_dims(value - gs.sum(row_sums, axis=-1), axis=-1)],
            axis=-1,
        )
        return gs.concatenate(
            [
                gs.concatenate(
                    [image_point, gs.expand_dims(row_sums, axis=-1)], axis=-1
                ),
                gs.expand_dims(last_row, axis=-2),
            ],
            axis=-2,
        )

    def inverse(self, image_point):
        r"""Inverse diffeomorphism at image point.

        :math:`f^{-1}: N \rightarrow M`

        Parameters
        ----------
        image_point : array-like, shape=[..., n-1, n-1]
            Image point.

        Returns
        -------
        base_point : array-like, shape=[..., n, n]
            Base point.
        """
        return self._concatenate_row_sums(image_point, self.value)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        df_p is a linear map from T_pM to T_f(p)N.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image of the base point.
        """
        out = self(tangent_vec)
        return repeat_out_multiple_ndim(
            out,
            self._space_ndim,
            (tangent_vec, base_point),
            self._image_space_ndim,
            (image_point,),
            out_ndim=self._image_space_ndim,
        )

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        df^-1_p is a linear map from T_f(p)N to T_pM

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        """
        out = self._concatenate_row_sums(image_tangent_vec, 0.0)
        return repeat_out_multiple_ndim(
            out,
            self._image_space_ndim,
            (image_tangent_vec, image_point),
            self._space_ndim,
            (base_point,),
            out_ndim=self._space_ndim,
        )


class NullRowSumsSymmetricMatrices(LevelSet, DiffeomorphicMatrixVectorSpace):
    r"""Space of null-row-sums symmetric matrices.

    Set of symmetric matrices with null row sums:

    .. math::

        \operatorname{Row_0}(n) = \{S \in \operatorname{Sym}(n)
        \mid S \mathbb{1}=0 \}

    Check out chapter 8 of [T2022]_ and [T2023]_ for more details.

    Parameters
    ----------
    n : int
        Integer representing the shapes of the matrices: n x n.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    .. [T2023] Thanwerdas, Yann. “Permutation-Invariant Log-Euclidean Geometries
        on Full-Rank Correlation Matrices,”
        November 2023. https://hal.science/hal-03878729.
    """

    def __init__(self, n, equip=True):
        self.n = n
        image_space = SymmetricMatrices(n - 1, equip=False)
        super().__init__(
            dim=image_space.dim,
            diffeo=ConstantValueRowSumsDiffeo(value=0.0),
            image_space=image_space,
            shape=(n, n),
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return MatricesMetric

    def _define_embedding_space(self):
        """Define embedding space of the manifold.

        Returns
        -------
        embedding_space : Manifold
            Instance of Manifold.
        """
        return SymmetricMatrices(n=self.n)

    def submersion(self, point):
        """Submersion that defines the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]

        Returns
        -------
        submersed_point : array-like, shape=[..., n]
        """
        return gs.sum(point, axis=-1)

    def tangent_submersion(self, vector, point):
        """Tangent submersion.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
        point : Ignored.

        Returns
        -------
        submersed_vector : array-like, shape=[..., n]
        """
        out = self.submersion(vector)
        return repeat_out(self.point_ndim, out, vector, point, out_shape=(self.n,))

    def _create_basis(self):
        """Compute the basis of the vector space."""
        indices, values = [], []
        k = -1
        for row in range(self.n - 1):
            for col in range(row, self.n - 1):
                k += 1
                if row == col:
                    indices.append((k, row, row))
                    values.append(1.0)
                else:
                    indices.extend([(k, row, col), (k, col, row)])
                    values.extend([1.0, 1.0])

        pre_basis = gs.array_from_sparse(indices, values, (k + 1, self.n, self.n))
        return self.matrix_representation(
            self.basis_representation(
                pre_basis,
            )
        )


class NullRowSumsPermutationInvariantMetric(EuclideanMetric):
    r"""A permutation-invariant metric on the space of null-row-sums symmetric matrices.

    It is flat Riemannian metric invariant by the congruence action
    of permutation matrices defined over a matrix vector space.

    Its associated quadratic form is:

    .. math::

        q(Y)=\alpha \operatorname{tr}\left(Y^2\right)
        +\delta \operatorname{tr}\left(\operatorname{Diag}(Y)^2\right)
        +\zeta \operatorname{tr}(Y)^2

    Check out chapter 8 of [T2022]_ and [T2023]_ for more details.

    Parameters
    ----------
    space : Manifold
    alpha : float
        Scalar multiplying first term of quadratic form.
    delta : float
        Scalar multiplying second term of quadratic form.
    zeta : float
        Scalar multiplying third term of quadratic form.

    References
    ----------
    .. [T2022] Yann Thanwerdas. Riemannian and stratified
        geometries on covariance and correlation matrices. Differential
        Geometry [math.DG]. Université Côte d'Azur, 2022.
    .. [T2023] Thanwerdas, Yann. “Permutation-Invariant Log-Euclidean Geometries
        on Full-Rank Correlation Matrices,”
        November 2023. https://hal.science/hal-03878729.
    """

    def __init__(self, space, alpha=None, delta=None, zeta=1.0):
        if alpha is None:
            alpha = 1.0 if space.n > 3 else 0.0

        if delta is None:
            delta = 1.0 if space.n > 2 else 0.0

        self._check_params(space, alpha, delta, zeta)
        super().__init__(space=space)
        self.alpha = alpha
        self.delta = delta
        self.zeta = zeta

    @staticmethod
    def _check_params(space, alpha, delta, zeta):
        r"""Check parameters of quadratic form.

        The following conditions must verify:
        - n > 3: :math:`\alpha>0, n\alpha+(n-2)\delta>0, n\alpha+(n-1)(\delta+n\zeta)>0`
        - n = 3: :math:`\alpha=0, \delta > 0, \delta + 3 \zeta > 0`
        - n = 2: :math:`\alpha=\delta=0, \zeta > 0`
        """
        n = space.n
        if n == 2:
            if alpha > gs.atol or delta > gs.atol or zeta < gs.atol:
                raise ValueError(
                    f"When n==2: alpha ({alpha}) and delta ({delta}) must be 0, "
                    f"and zeta ({zeta}) > 0. "
                    "Check thanwerdas2022 theorem 8.7"
                )
            return

        elif n == 3:
            cond = delta + 3 * zeta
            if alpha > gs.atol or delta < gs.atol or cond < gs.atol:
                raise ValueError(
                    f"When n==3: alpha ({alpha}) must be 0, delta ({delta}) > 0 "
                    f"and an inequality greater than 0: {cond}. "
                    "Check thanwerdas2022 theorem 8.7"
                )
            return

        cond_1 = n * alpha + (n - 2) * delta
        cond_2 = n * alpha + (n - 1) * (delta + n * zeta)
        if cond_1 < gs.atol or cond_2 < gs.atol:
            raise ValueError(
                f"Inequalities should be greater than 0, but: {cond_1} and {cond_2}. "
                "Check thanwerdas2022 theorem 8.7"
            )

    def _quadratic_form(self, tangent_vec):
        """Quadratic form associated to inner product.

        Parameters
        ----------
        tangent_vec: array-like, shape=[..., n, n]
            Tangent vector at base point.
        """
        if self.alpha > gs.atol:
            comp = gs.matmul(tangent_vec, tangent_vec)
            out_alpha = self.alpha * gs.trace(comp)
        else:
            out_alpha = 0.0
        out_delta = (
            self.delta * gs.sum(Matrices.diagonal(tangent_vec) ** 2, axis=-1)
            if self.delta > gs.atol
            else 0.0
        )
        out_zeta = self.zeta * gs.trace(tangent_vec) ** 2

        return out_alpha + out_delta + out_zeta

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a: array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b: array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point: array-like, shape=[..., n, n]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        inner_prod = (1 / 2) * (
            self._quadratic_form(tangent_vec_a + tangent_vec_b)
            - self._quadratic_form(tangent_vec_a)
            - self._quadratic_form(tangent_vec_b)
        )
        return repeat_out(
            self._space.point_ndim, inner_prod, tangent_vec_a, tangent_vec_b, base_point
        )

    def squared_norm(self, vector, base_point=None):
        """Compute the square of the norm of a vector.

        Squared norm of a vector associated to the inner product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
            Vector.
        base_point : array-like, shape=[..., n, n]
            Base point.
            Optional, default: None.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm.
        """
        return self._quadratic_form(vector)



import pytest

import geomstats.backend as gs
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase


class ComplexRiemannianMetricTestCase(RiemannianMetricTestCase):
    @pytest.mark.random
    def test_inner_product_is_symmetric(self, n_points, atol):
        """Check that the inner product is Hermitian."""
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        inner_product_ab = self.space.metric.inner_product(
            tangent_vec_a, tangent_vec_b, base_point
        )
        inner_product_ba = self.space.metric.inner_product(
            tangent_vec_b, tangent_vec_a, base_point
        )
        self.assertAllClose(inner_product_ab, gs.conj(inner_product_ba), atol=atol)

    @pytest.mark.type
    def test_inner_product_is_complex(self, n_points):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        result = self.space.metric.inner_product(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertTrue(gs.is_complex(result))

    @pytest.mark.type
    def test_dist_is_real(self, n_points):
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        result = self.space.metric.dist(point_a, point_b)
        self.assertTrue(not gs.is_complex(result))

    @pytest.mark.type
    def test_log_is_complex(self, n_points):
        base_point = self.data_generator.random_point(n_points)
        point = self.data_generator.random_point(n_points)

        result = self.space.metric.log(point, base_point)
        self.assertTrue(gs.is_complex(result))

    @pytest.mark.type
    def test_exp_is_complex(self, n_points):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        result = self.space.metric.exp(tangent_vec, base_point)
        self.assertTrue(gs.is_complex(result))



import geomstats.backend as gs
from geomstats.geometry.diffeo import AutodiffDiffeo
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase


class LowerMatrixLog(AutodiffDiffeo):
    """Matrix logarithm diffeomorphism.

    An alternative implementation for testing
    purposes.
    """

    @staticmethod
    def __call__(base_point):
        """Compute the matrix log.

        Parameters
        ----------
        base_point : array_like, shape=[..., n, n]
            Symmetric matrix.

        Returns
        -------
        log : array_like, shape=[..., n, n]
            Matrix logarithm of base_point.
        """
        return gs.linalg.logm(base_point)

    @staticmethod
    def inverse(image_point):
        """Compute the matrix exponential.

        Parameters
        ----------
        image_point : array_like, shape=[..., n, n]
            Symmetric matrix.

        Returns
        -------
        exponential : array_like, shape=[..., n, n]
            Exponential of image_point.
        """
        return gs.linalg.expm(image_point)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image of the base point.
        """
        raise NotImplementedError("Not implemented.")


class CholeskyMetricTestCase(RiemannianMetricTestCase):
    def test_diag_inner_product(
        self, tangent_vec_a, tangent_vec_b, base_point, expected, atol
    ):
        res = self.space.metric.diag_inner_product(
            tangent_vec_a,
            tangent_vec_b,
            base_point,
        )
        self.assertAllClose(res, expected, atol=atol)

    def test_strictly_lower_inner_product(
        self, tangent_vec_a, tangent_vec_b, expected, atol
    ):
        res = self.space.metric.strictly_lower_inner_product(
            tangent_vec_a,
            tangent_vec_b,
        )
        self.assertAllClose(res, expected, atol=atol)



"""Brownian motion defined on a manifold."""

import geomstats.backend as gs


class BrownianMotion:
    """Class to generate a realization of Brownian motion on a manifold.

    Parameters
    ----------
    space : Manifold
        Manifold to generate Brownian motion on.

    Example
    --------
    >>> import os
    >>> os.environ["GEOMSTATS_BACKEND"] = "pytorch"
    >>> import geomstats.backend as gs
    >>> from geomstats.geometry.euclidean import Euclidean
    >>> from geomstats.distributions.brownian_motion import BrownianMotion
    >>> manifold = Euclidean(dim=3)
    >>> euclidean_brownian_motion = BrownianMotion(manifold)
    >>> sample_path = euclidean_brownian_motion.sample_path(
            end_time=1,
            n_steps=50,
            initial_point=gs.array([0.0, 0.0, 0.0])
        )

    References
    ----------
    .. [H2022] Elton P. Hsu,
        "Stochastic Analysis On Manifolds",
        American Mathematical Soc. (2002): 71-99.
    """

    def __init__(self, space):
        self.space = space
        self._check_coordinates(space)

    def _check_coordinates(self, space):
        """Check the manifold is defined in intrinsic coordinates."""
        if not space.intrinsic:
            raise ValueError(
                "Space should be equipped with intrinsic coordinates to create Brownian"
                "motion over the local parametrization."
            )

    def sample_path(self, end_time, n_steps, initial_point):
        """Generate a sample path of Brownian motion.

        Parameters
        ----------
        end_time : float
            Final time of the path.
        n_steps : int
            Number of steps in the path.
        initial_point : array-like, shape=[..., dim]
            Initial point of the path at time 0.

        Returns
        -------
        path : array-like, shape=[..., n_steps, dim]
            Sample path of Brownian motion.

        """
        step_size = end_time / n_steps

        paths = [initial_point]
        for _ in range(1, n_steps):
            paths.append(self._step(step_size, paths[-1]))
        return gs.stack(paths)

    def _step(self, step_size, current_point):
        """Calulate one increment of a Brownian motion path.

        Parameters
        ----------
        step_size : float
            Size of the step to be taken in the Brownian motion.
        current_point : array-like, shape=[..., dim]
            Current point in the Brownian motion path.

        Returns
        -------
        next_point : array-like, shape=[..., dim]
            Next point in the Brownian motion path after taking the step.

        Notes
        -----
        Method uses the Euler-Maruyama integration scheme.
        Brownian motion is described in the Ito form in intrinsic coordinates as a
        stochastic differential equation, [H2022] (example 3.3.5).

        """
        sigma = gs.linalg.sqrtm(self.space.metric.cometric_matrix(current_point))
        christoffels = self.space.metric.christoffels(current_point)
        cometric_matrix = self.space.metric.cometric_matrix(current_point)
        drift = (
            -0.5
            * gs.einsum("...klm,...lm->...k", christoffels, cometric_matrix)
            * step_size
        )
        batch_shape = current_point.shape[: -self.space.point_ndim]
        diffusion = gs.einsum(
            "...ij,...j->...i",
            sigma,
            gs.random.normal(size=batch_shape + (self.space.dim,)) * gs.sqrt(step_size),
        )

        return current_point + drift + diffusion



"""Diffeomorpism implementation."""

import abc
import math

import geomstats.backend as gs
from geomstats.vectorization import get_batch_shape, repeat_out_multiple_ndim


class Diffeo:
    r"""Diffeormorphism.

    Let :math:`f` be the diffeomorphism
    :math:`f: M \rightarrow N` of the manifold
    :math:`M` into the manifold :math:`N`.
    """

    @abc.abstractmethod
    def __call__(self, base_point):
        """Diffeomorphism at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., *space_shape]
            Base point.

        Returns
        -------
        image_point : array-like, shape=[..., *image_shape]
            Image point.
        """

    @abc.abstractmethod
    def inverse(self, image_point):
        r"""Inverse diffeomorphism at image point.

        :math:`f^{-1}: N \rightarrow M`

        Parameters
        ----------
        image_point : array-like, shape=[..., *image_shape]
            Image point.

        Returns
        -------
        base_point : array-like, shape=[..., *space_shape]
            Base point.
        """

    @abc.abstractmethod
    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        df_p is a linear map from T_pM to T_f(p)N.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image of the base point.

        Notes
        -----
        Signature choice (i.e. having the possibility to pass both base and image
        points) comes from performance considerations.
        In several methods (e.g. `PullbackDiffeoMetric.inner_product`) the need
        to call `tangent_diffeomorphism` comes after having called
        `diffeomorphism`, which means we have both `base_point` and
        `image_point` available.
        In some cases, `tangent_diffeomorphism` needs access to `base_point`
        (e.g. `SRVTransform`), while in others, it needs access to `image_point`
        (e.g. `CholeskyMap`).
        By passing both, we give the corresponding implementation the possibility
        to choose the one that is more convenient (performancewise).
        If we pass only one of the two, it has the mechanims to perform the
        necessary transformations (e.g. calling `diffeomorphism` or
        `inverse_diffeomorphism` on it).
        """

    @abc.abstractmethod
    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        df^-1_p is a linear map from T_f(p)N to T_pM

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.

        Notes
        -----
        See `tangent_diffeomorphism` docstrings for signature considerations.
        """


class AutodiffDiffeo(Diffeo):
    """Diffeomorphism through autodiff."""

    def __init__(self, space_shape, image_space_shape=None):
        super().__init__()
        if image_space_shape is None:
            image_space_shape = space_shape

        self._space_shape = space_shape
        self._space_point_ndim = len(space_shape)

        self._image_space_shape = image_space_shape
        self._image_space_point_ndim = len(image_space_shape)

        self._shape_prod = math.prod(space_shape)
        self._image_shape_prod = math.prod(image_space_shape)

    def jacobian(self, base_point):
        r"""Jacobian of the diffeomorphism at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., *space_shape]
            Base point.

        Returns
        -------
        mat : array-like, shape=[..., *image_shape, *space_shape]
            Jacobian of the diffeomorphism.
        """
        return gs.autodiff.jacobian_vec(
            self.__call__, point_ndim=self._space_point_ndim
        )(base_point)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image of the base point.
        """
        if base_point is None:
            base_point = self.inverse(image_point)

        batch_shape = get_batch_shape(self._space_point_ndim, tangent_vec, base_point)
        flat_batch_shape = (-1,) if batch_shape else ()

        j_flat = gs.reshape(
            self.jacobian(base_point),
            flat_batch_shape + (self._image_shape_prod, self._shape_prod),
        )
        tv_flat = gs.reshape(tangent_vec, flat_batch_shape + (self._shape_prod,))

        image_tv = gs.reshape(
            gs.einsum("...ij,...j->...i", j_flat, tv_flat),
            batch_shape + self._image_space_shape,
        )

        return image_tv

    def inverse_jacobian(self, image_point):
        r"""Jacobian of the inverse diffeomorphism at image point.

        Parameters
        ----------
        image_point : array-like, shape=[..., *image_shape]
            Base point.

        Returns
        -------
        mat : array-like, shape=[..., *shape, *image_shape]
            Jacobian of the inverse diffeomorphism.
        """
        return gs.autodiff.jacobian_vec(
            self.inverse, point_ndim=self._image_space_point_ndim
        )(image_point)

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Tangent diffeomorphism at image point.

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Tangent vector at base point.
        image_point : array-like, shape=[..., *image_shape]
            Base point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        """
        if image_point is None:
            image_point = self(base_point)

        batch_shape = get_batch_shape(
            self._image_space_point_ndim, image_tangent_vec, image_point
        )
        flat_batch_shape = (-1,) if batch_shape else ()

        j_flat = gs.reshape(
            self.inverse_jacobian(image_point),
            flat_batch_shape + (self._shape_prod, self._image_shape_prod),
        )

        itv_flat = gs.reshape(
            image_tangent_vec, flat_batch_shape + (self._image_shape_prod,)
        )

        tv = gs.reshape(
            gs.einsum("...ij,...j->...i", j_flat, itv_flat),
            batch_shape + self._space_shape,
        )
        return tv


class ReversedDiffeo(Diffeo):
    """Reverses the direction of a diffeomorphism.

    Parameters
    ----------
    diffeo : Diffeo.
    """

    def __init__(self, diffeo):
        self.diffeo = diffeo

    def __call__(self, base_point):
        """Diffeomorphism at base point."""
        return self.diffeo.inverse(base_point)

    def inverse(self, image_point):
        """Inverse diffeomorphism at image point."""
        return self.diffeo(image_point)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        """Tangent diffeomorphism at base point."""
        return self.diffeo.inverse_tangent(
            tangent_vec, image_point=base_point, base_point=image_point
        )

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        """Tangent diffeomorphism at image point."""
        return self.diffeo.tangent(
            image_tangent_vec, base_point=image_point, image_point=image_point
        )


class ComposedDiffeo(Diffeo):
    """A composed diffeomorphism.

    Parameters
    ----------
    diffeos : list[Diffeo]
        An (ordered) list of diffeomorphisms.
    """

    def __init__(self, diffeos):
        self.diffeos = diffeos

    def __call__(self, base_point):
        """Diffeomorphism at base point."""
        image_point = base_point
        for diffeo in self.diffeos:
            image_point = diffeo(image_point)

        return image_point

    def inverse(self, image_point):
        """Inverse diffeomorphism at image point."""
        base_point = image_point
        for diffeo in reversed(self.diffeos):
            base_point = diffeo.inverse(base_point)
        return base_point

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        """Tangent diffeomorphism at base point."""
        if base_point is None:
            base_point = self.inverse(image_point)

        for diffeo in self.diffeos:
            image_point = diffeo(base_point)
            image_tangent_vec = diffeo.tangent(
                tangent_vec, base_point=base_point, image_point=image_point
            )
            tangent_vec = image_tangent_vec
            base_point = image_point

        return image_tangent_vec

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        """Inverse tangent diffeomorphism at image point."""
        if image_point is None:
            image_point = self(base_point)

        for diffeo in reversed(self.diffeos):
            base_point = diffeo.inverse(image_point)
            tangent_vec = diffeo.inverse_tangent(
                image_tangent_vec, image_point=image_point, base_point=base_point
            )
            image_tangent_vec = tangent_vec
            image_point = base_point

        return tangent_vec


class VectorSpaceDiffeo(Diffeo):
    """A diffeo between vector spaces."""

    def __init__(self, space_ndim, image_space_ndim=None):
        super().__init__()
        self.space_ndim = space_ndim
        self.image_space_ndim = image_space_ndim or space_ndim

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        df_p is a linear map from T_pM to T_f(p)N.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image of the base point.
        """
        out = self(tangent_vec)
        return repeat_out_multiple_ndim(
            out,
            self.space_ndim,
            (tangent_vec, base_point),
            self.image_space_ndim,
            (image_point,),
            out_ndim=self.image_space_ndim,
        )

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        df^-1_p is a linear map from T_f(p)N to T_pM

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        """
        out = self.inverse(image_tangent_vec)
        return repeat_out_multiple_ndim(
            out,
            self.image_space_ndim,
            (image_tangent_vec, image_point),
            self.space_ndim,
            (base_point,),
            out_ndim=self.space_ndim,
        )


class InvolutionDiffeomorphism(Diffeo):
    """A diffeomorphism that is also an involution."""

    def inverse(self, image_point):
        r"""Inverse diffeomorphism at image point.

        :math:`f^{-1}: N \rightarrow M`

        Parameters
        ----------
        image_point : array-like, shape=[..., *image_shape]
            Image point.

        Returns
        -------
        base_point : array-like, shape=[..., *space_shape]
            Base point.
        """
        return self(image_point)

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        df^-1_p is a linear map from T_f(p)N to T_pM

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., *image_shape]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., *image_shape]
            Image point.
        base_point : array-like, shape=[..., *space_shape]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *space_shape]
            Tangent vector at base point.
        """
        return self.tangent(
            image_tangent_vec, base_point=image_point, image_point=base_point
        )



"""Complex Hermitian space.

Lead author: Yann Cabanes.
"""

import geomstats.backend as gs
from geomstats.geometry.base import ComplexVectorSpace
from geomstats.geometry.complex_riemannian_metric import ComplexRiemannianMetric


class Hermitian(ComplexVectorSpace):
    """Class for Hermitian spaces.

    By definition, a Hermitian space is a complex vector space
    of a given dimension, equipped with a Hermitian metric.

    Parameters
    ----------
    dim : int
        Dimension of the Hermitian space.
    """

    def __init__(self, dim, equip=True):
        super().__init__(shape=(dim,), equip=equip)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return HermitianMetric

    @property
    def identity(self):
        """Identity of the group.

        Returns
        -------
        identity : array-like, shape=[n]
        """
        return gs.zeros(self.dim, dtype=gs.get_default_cdtype())

    def exp(self, tangent_vec, base_point=None):
        """Compute the group exponential, which is simply the addition.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n]
            Point from which the exponential is computed.

        Returns
        -------
        point : array-like, shape=[..., n]
            Group exponential.
        """
        return tangent_vec + base_point


class HermitianMetric(ComplexRiemannianMetric):
    """Class for Hermitian metrics.

    As a Riemannian metric, the Hermitian metric is:

    - flat: the inner-product is independent of the base point.
    - positive definite: it has signature (dimension, 0, 0),
      where dimension is the dimension of the Hermitian space.
    """

    def metric_matrix(self, base_point=None):
        """Compute the inner-product matrix, independent of the base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_prod_mat : array-like, shape=[..., dim, dim]
            Inner-product matrix.
        """
        mat = gs.eye(self._space.dim, dtype=gs.get_default_cdtype())
        if base_point is not None and base_point.ndim > 1:
            return gs.broadcast_to(mat, base_point.shape + (self._space.dim,))
        return mat

    @staticmethod
    def inner_product(tangent_vec_a, tangent_vec_b, base_point=None):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a: array-like, shape=[..., dim]
            Tangent vector at base point.
        tangent_vec_b: array-like, shape=[..., dim]
            Tangent vector at base point.
        base_point: array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        return gs.dot(gs.conj(tangent_vec_a), tangent_vec_b)

    @staticmethod
    def norm(vector, base_point=None):
        """Compute norm of a vector.

        Norm of a vector associated to the inner product
        at the tangent space at a base point.

        Note: This only works for positive-definite
        Riemannian metrics and inner products.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        norm : array-like, shape=[...,]
            Norm.
        """
        return gs.linalg.norm(vector, axis=-1)

    @staticmethod
    def exp(tangent_vec, base_point):
        """Compute exp map of a base point in tangent vector direction.

        The Riemannian exponential is vector addition in the Hermitian space.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at base point.
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., dim]
            Riemannian exponential.
        """
        return base_point + tangent_vec

    @staticmethod
    def log(point, base_point):
        """Compute log map using a base point and other point.

        The Riemannian logarithm is the subtraction in the Hermitian space.

        Parameters
        ----------
        point: array-like, shape=[..., dim]
            Point.
        base_point: array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        log: array-like, shape=[..., dim]
            Riemannian logarithm.
        """
        return point - base_point



from ._dispatch import _common
from ._dispatch import numpy as _np

_modify_func_default_dtype = _common._modify_func_default_dtype
_allow_complex_dtype = _common._allow_complex_dtype


rand = _modify_func_default_dtype(
    copy=False, kw_only=True, target=_allow_complex_dtype(target=_np.random.rand)
)

uniform = _modify_func_default_dtype(
    copy=False, kw_only=True, target=_allow_complex_dtype(target=_np.random.uniform)
)


normal = _modify_func_default_dtype(
    copy=False, kw_only=True, target=_allow_complex_dtype(target=_np.random.normal)
)

multivariate_normal = _modify_func_default_dtype(
    copy=False,
    kw_only=True,
    target=_allow_complex_dtype(target=_np.random.multivariate_normal),
)


def choice(*args, **kwargs):
    return _np.random.default_rng().choice(*args, **kwargs)



"""Mixin for manifolds of probability distributions."""

import math

import geomstats.backend as gs
from geomstats.vectorization import get_batch_shape


class InformationManifoldMixin:
    """Mixin for manifolds of probability distributions."""

    def __init__(self, support_shape, **kwargs):
        self.support_shape = support_shape
        super().__init__(**kwargs)

    def sample(self, point, n_samples=1):
        """Sample from the probability distribution.

        Sample from the probability distribution with parameters provided
        by point. This gives n_samples points.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point representing a probability distribution.
        n_samples : int
            Number of points to sample for each set of parameters in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples]
            Sample from the probability distributions.
        """
        raise NotImplementedError("The sample method is not yet implemented.")

    def point_to_pdf(self, point):
        """Compute pdf associated to point.

        Compute the probability density function of the probability
        distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point representing a probability distribution.

        Returns
        -------
        pdf : function
            Probability density function of the probability distribution with
            parameters provided by point.
        """
        raise NotImplementedError("`point_to_pdf` has not yet been implemented.")

    def point_to_cdf(self, point):
        """Compute cdf associated to point.

        Compute the cumulative density function of the probability
        distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point representing a probability distribution.

        Returns
        -------
        cdf : function
            Cumulative density function of the probability distribution with
            parameters provided by point.
        """
        raise NotImplementedError("`point_to_cdf` has not yet been implemented.")


class ScipyRandomVariable:
    """A random variable."""

    def __init__(self, space, scp_rvs, scp_pdf=None):
        self.space = space
        self.scp_rvs = scp_rvs
        self.scp_pdf = scp_pdf


class ScipyUnivariateRandomVariable(ScipyRandomVariable):
    """A univariate random variable."""

    @staticmethod
    def _unflatten_res(flat_res, expected_shape):
        return gs.reshape(flat_res, expected_shape)

    def rvs(self, point, n_samples=1):
        """Sample from a univariate distribution.

        Parameters
        ----------
        point : array-like, shape=[..., *space.shape]
            Point representing a univariate distribution.
        n_samples : int
            Number of points to sample with each pair of parameters in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples, *support_shape]
            Sample from distribution.
        """
        batch_shape = get_batch_shape(self.space.point_ndim, point)
        n_points = math.prod(batch_shape)

        pre_flat_shape = batch_shape + (n_samples,)
        params = self._flatten_params(point, pre_flat_shape)

        size = n_points * n_samples
        flat_sample = gs.from_numpy(self.scp_rvs(**params, size=size))

        expected_shape = batch_shape + (n_samples,) + self.space.support_shape
        return self._unflatten_res(flat_sample, expected_shape)

    def pdf(self, sample, point):
        """Evaluate the probability density function at x.

        Parameters
        ----------
        x : array-like, shape=[n_samples, *support_shape]
            Sample points at which to compute the probability density function.
        point : array-like, shape=[..., *space.shape]
            Point representing a distribution.

        Returns
        -------
        pdf_at_x : array-like, shape=[..., n_samples]
            Values of pdf at x for each value of the parameters provided
            by point.
        """
        batch_shape = get_batch_shape(self.space.point_ndim, point)
        n_points = math.prod(batch_shape)
        n_samples = sample.shape[0]

        pre_flat_shape = batch_shape + (n_samples,)
        params = self._flatten_params(point, pre_flat_shape)

        sample_ = gs.reshape(gs.broadcast_to(sample, (n_points, n_samples)), (-1,))

        pdf = gs.from_numpy(self.scp_pdf(sample_, **params))

        expected_shape = batch_shape + (n_samples,) + self.space.support_shape
        return self._unflatten_res(pdf, expected_shape)


class ScipyMultivariateRandomVariable(ScipyRandomVariable):
    """A multivariate random variable."""

    def _rvs_single(self, point, n_samples):
        return gs.to_ndarray(
            gs.from_numpy(self.scp_rvs(point, size=n_samples)),
            to_ndim=len(self.space.support_shape) + 1,
        )

    def rvs(self, point, n_samples=1):
        """Sample from a multivariate distribution.

        Parameters
        ----------
        point : array-like, shape=[..., *space.shape]
            Point representing a distribution.
        n_samples : int
            Number of points to sample with each pair of parameters in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples, *support_shape]
            Sample from distribution.
        """
        if point.ndim > self.space.point_ndim:
            return gs.stack([self._rvs_single(point_, n_samples) for point_ in point])

        return self._rvs_single(point, n_samples)

    def _pdf_single(self, x, point):
        out = self.scp_pdf(x, point)
        if x.shape[0] == 1 and point.ndim == self.space.point_ndim:
            return gs.array([out])
        return gs.from_numpy(out)

    def pdf(self, x, point):
        """Evaluate the probability density function at x.

        Parameters
        ----------
        x : array-like, shape=[n_samples, *support_shape]
            Sample points at which to compute the probability density function.
        point : array-like, shape=[..., *space.shape]
            Point representing a distribution.

        Returns
        -------
        pdf_at_x : array-like, shape=[..., n_samples, *support_shape]
            Values of pdf at x for each value of the parameters provided
            by point.
        """
        if point.ndim > self.space.point_ndim:
            return gs.stack([self._pdf_single(x, point_) for point_ in point])

        return self._pdf_single(x, point)



"""The Visualization Package."""

import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # NOQA

import geomstats.backend as gs
from geomstats.visualization.hyperbolic import (
    KleinDisk,
    PoincareDisk,
    PoincareHalfPlane,
)
from geomstats.visualization.hypersphere import Circle, Sphere
from geomstats.visualization.poincare_polydisk import PoincarePolyDisk
from geomstats.visualization.pre_shape import KendallDisk, KendallSphere
from geomstats.visualization.spd_matrices import Ellipses
from geomstats.visualization.special_euclidean import SpecialEuclidean2
from geomstats.visualization.special_orthogonal import (
    Arrow3D,
    Trihedron,
    convert_to_trihedron,
)

AX_SCALE = 1.2


IMPLEMENTED = [
    "SO3_GROUP",
    "SE3_GROUP",
    "SE2_GROUP",
    "S1",
    "S2",
    "H2_poincare_disk",
    "H2_poincare_half_plane",
    "H2_klein_disk",
    "poincare_polydisk",
    "S32",
    "M32",
    "S33",
    "M33",
    "SPD2",
]


def tutorial_matplotlib():
    """Configure style for matplotlib tutorial."""
    fontsize = 12
    matplotlib.rc("font", size=fontsize)
    matplotlib.rc("text")
    matplotlib.rc("legend", fontsize=fontsize)
    matplotlib.rc("axes", titlesize=21, labelsize=14)
    matplotlib.rc(
        "font",
        family="times",
        serif=["Computer Modern Roman"],
        monospace=["Computer Modern Typewriter"],
    )


def plot(points, ax=None, space=None, coords_type=None, **point_draw_kwargs):
    """Plot points in one of the implemented manifolds.

    The implemented manifolds are:

    - the special orthogonal group SO(3)
    - the special Euclidean group SE(3)
    - the circle S1 and the sphere S2
    - the hyperbolic plane (the Poincare disk, the Poincare
      half plane and the Klein disk)
    - the Poincare polydisk
    - the Kendall shape space of 2D triangles
    - the Kendall shape space of 3D triangles

    Parameters
    ----------
    points : array-like, shape=[..., dim]
        Points to be plotted.
    space: str, optional, {'SO3_GROUP', 'SE3_GROUP', 'S1', 'S2',
        'H2_poincare_disk', 'H2_poincare_half_plane', 'H2_klein_disk',
        'poincare_polydisk', 'S32', 'M32', 'S33', 'M33', 'SPD2'}
    coords_type: str, optional, {'extrinsic', 'ball', 'half-space', 'pre-shape'}
    """
    if space not in IMPLEMENTED:
        raise NotImplementedError(
            "The plot function is not implemented"
            " for space {}. The spaces available for visualization"
            " are: {}.".format(space, IMPLEMENTED)
        )

    if points is None:
        raise ValueError("No points given for plotting.")

    if points.ndim < 2:
        points = gs.expand_dims(points, 0)

    if space in ("SO3_GROUP", "SE3_GROUP"):
        if ax is None:
            ax = plt.subplot(111, projection="3d")
        if space == "SE3_GROUP":
            ax_s = AX_SCALE * gs.amax(gs.abs(points[:, 3:6]))
        elif space == "SO3_GROUP":
            ax_s = AX_SCALE * gs.amax(gs.abs(points[:, :3]))
        ax_s = float(ax_s)
        bounds = (-ax_s, ax_s)
        plt.setp(
            ax,
            xlim=bounds,
            ylim=bounds,
            zlim=bounds,
            xlabel="X",
            ylabel="Y",
            zlabel="Z",
        )
        trihedrons = convert_to_trihedron(points, space=space)
        for t in trihedrons:
            t.draw(ax, **point_draw_kwargs)

    elif space == "S1":
        circle = Circle()
        ax = circle.set_ax(ax=ax)
        circle.add_points(points)
        circle.draw(ax, **point_draw_kwargs)

    elif space == "S2":
        sphere = Sphere()
        ax = sphere.set_ax(ax=ax)
        sphere.add_points(points)
        sphere.draw(ax, **point_draw_kwargs)

    elif space == "H2_poincare_disk":
        if coords_type is None:
            coords_type = "extrinsic"
        poincare_disk = PoincareDisk(coords_type=coords_type)
        ax = poincare_disk.set_ax(ax=ax)
        poincare_disk.add_points(points)
        poincare_disk.draw(ax, **point_draw_kwargs)
        plt.axis("off")

    elif space == "poincare_polydisk":
        if coords_type is None:
            coords_type = "extrinsic"
        n_disks = points.shape[1]
        poincare_poly_disk = PoincarePolyDisk(coords_type=coords_type, n_disks=n_disks)
        n_columns = int(gs.ceil(n_disks**0.5))
        n_rows = int(gs.ceil(n_disks / n_columns))

        axis_list = []

        for i_disk in range(n_disks):
            axis_list.append(ax.add_subplot(n_rows, n_columns, i_disk + 1))

        for i_disk, one_ax in enumerate(axis_list):
            ax = poincare_poly_disk.set_ax(ax=one_ax)
            poincare_poly_disk.clear_points()
            poincare_poly_disk.add_points(points[:, i_disk, ...])
            poincare_poly_disk.draw(ax, **point_draw_kwargs)

    elif space == "H2_poincare_half_plane":
        if coords_type is None:
            coords_type = "half-space"
        poincare_half_plane = PoincareHalfPlane(coords_type=coords_type)
        ax = poincare_half_plane.set_ax(ax=ax)
        poincare_half_plane.add_points(points)
        poincare_half_plane.draw(ax, **point_draw_kwargs)

    elif space == "H2_klein_disk":
        klein_disk = KleinDisk()
        ax = klein_disk.set_ax(ax=ax)
        klein_disk.add_points(points)
        klein_disk.draw(ax, **point_draw_kwargs)

    elif space == "SE2_GROUP":
        plane = SpecialEuclidean2()
        ax = plane.set_ax(ax=ax)
        plane.add_points(points)
        plane.draw_points(ax, **point_draw_kwargs)

    elif space == "S32":
        sphere = KendallSphere()
        sphere.add_points(points)
        sphere.draw()
        sphere.draw_points()
        ax = sphere.ax

    elif space == "M32":
        sphere = KendallSphere(coords_type="extrinsic")
        sphere.add_points(points)
        sphere.draw()
        sphere.draw_points()
        ax = sphere.ax

    elif space == "S33":
        disk = KendallDisk()
        disk.add_points(points)
        disk.draw()
        disk.draw_points()
        ax = disk.ax

    elif space == "M33":
        disk = KendallDisk(coords_type="extrinsic")
        disk.add_points(points)
        disk.draw()
        disk.draw_points()
        ax = disk.ax

    elif space == "SPD2":
        ellipses = Ellipses()
        ellipses.draw_points(points=points)

    return ax



"""The manifold of lower triangular matrices with positive diagonal elements.

Lead authors: Olivier Bisson and Saiteja Utpala.

References
----------
.. [T2022] Yann Thanwerdas. Riemannian and stratified
    geometries on covariance and correlation matrices. Differential
    Geometry [math.DG]. Université Côte d'Azur, 2022.
"""

import geomstats.backend as gs
from geomstats.geometry.base import DiffeomorphicManifold, LevelSet, VectorSpaceOpenSet
from geomstats.geometry.diffeo import Diffeo
from geomstats.geometry.invariant_metric import _InvariantMetricMatrix
from geomstats.geometry.lie_group import MatrixLieGroup
from geomstats.geometry.lower_triangular_matrices import LowerTriangularMatrices
from geomstats.geometry.matrices import Matrices, MatricesMetric
from geomstats.geometry.open_hemisphere import OpenHemisphere, OpenHemispheresProduct
from geomstats.geometry.pullback_metric import PullbackDiffeoMetric
from geomstats.geometry.riemannian_metric import RiemannianMetric


class PositiveLowerTriangularMatrices(MatrixLieGroup, VectorSpaceOpenSet):
    """Manifold of lower triangular matrices with >0 diagonal.

    This manifold is also called the cholesky space.

    Parameters
    ----------
    n : int
        Integer representing the shape of the matrices: n x n.

    References
    ----------
    .. [TP2019] Z Lin. "Riemannian Geometry of Symmetric
        Positive Definite Matrices Via Cholesky Decomposition"
        SIAM journal on Matrix Analysis and Applications , 2019.
        https://arxiv.org/abs/1908.09326
    """

    def __init__(self, n, equip=True):
        super().__init__(
            representation_dim=n,
            lie_algebra=LowerTriangularMatrices(n, equip=False),
            dim=int(n * (n + 1) / 2),
            embedding_space=LowerTriangularMatrices(n, equip=False),
            equip=equip,
        )
        self.n = n

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return CholeskyMetric

    def random_point(self, n_samples=1, bound=1.0):
        """Sample in PLT.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Side of hypercube support of the uniform distribution.
            Optional, default: 1.0

        Returns
        -------
        point : array-like, shape=[..., *point_shape]
           Sample.
        """
        batch_shape = (n_samples,) if n_samples > 1 else ()
        shape = batch_shape + (self.dim - self.n,)

        lower_part = bound * (gs.random.rand(*shape) - 0.5) * 2

        diag_shape = batch_shape + (self.n,)
        diagonal = bound * gs.random.rand(*diag_shape) + gs.atol

        return gs.mat_from_diag_triu_tril(diagonal, gs.array([0.0]), lower_part)

    def belongs(self, point, atol=gs.atol):
        """Check if mat is lower triangular with >0 diagonal.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix to be checked.
        atol : float
            Tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean denoting if mat belongs to cholesky space.
        """
        is_lower_triangular = self.embedding_space.belongs(point, atol)
        diagonal = Matrices.diagonal(point)
        is_positive = gs.all(diagonal > 0, axis=-1)
        return gs.logical_and(is_lower_triangular, is_positive)

    def projection(self, point):
        """Project a matrix to the PLT space.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]

        Returns
        -------
        projected: array-like, shape=[..., n, n]
        """
        vec_diag = gs.abs(Matrices.diagonal(point))
        vec_diag = gs.where(vec_diag < gs.atol, gs.atol, vec_diag)
        diag = gs.vec_to_diag(vec_diag)
        strictly_lower_triangular = gs.tril(point, k=-1)
        return diag + strictly_lower_triangular


class CholeskyMetric(RiemannianMetric):
    """Class for Cholesky metric on Cholesky space.

    References
    ----------
    .. [TP2019] . "Riemannian Geometry of Symmetric
        Positive Definite Matrices Via Cholesky Decomposition"
        SIAM journal on Matrix Analysis and Applications , 2019.
        https://arxiv.org/abs/1908.09326
    """

    @staticmethod
    def diag_inner_product(tangent_vec_a, tangent_vec_b, base_point):
        """Compute the inner product using only diagonal elements.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        ip_diagonal : array-like, shape=[...]
            Inner-product.
        """
        inv_sqrt_diagonal = gs.power(Matrices.diagonal(base_point), -2)
        tangent_vec_a_diagonal = Matrices.diagonal(tangent_vec_a)
        tangent_vec_b_diagonal = Matrices.diagonal(tangent_vec_b)
        prod = tangent_vec_a_diagonal * tangent_vec_b_diagonal * inv_sqrt_diagonal
        return gs.sum(prod, axis=-1)

    @staticmethod
    def strictly_lower_inner_product(tangent_vec_a, tangent_vec_b):
        """Compute the inner product using only strictly lower triangular elements.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at base point.

        Returns
        -------
        ip_sl : array-like, shape=[...]
            Inner-product.
        """
        sl_tagnet_vec_a = gs.tril_to_vec(tangent_vec_a, k=-1)
        sl_tagnet_vec_b = gs.tril_to_vec(tangent_vec_b, k=-1)
        ip_sl = gs.dot(sl_tagnet_vec_a, sl_tagnet_vec_b)
        return ip_sl

    @classmethod
    def inner_product(cls, tangent_vec_a, tangent_vec_b, base_point):
        """Compute the inner product.

        Compute the inner-product of tangent_vec_a and tangent_vec_b
        at point base_point using the cholesky Riemannian metric.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        inner_product : array-like, shape=[...]
            Inner-product.
        """
        diag_inner_product = cls.diag_inner_product(
            tangent_vec_a, tangent_vec_b, base_point
        )
        strictly_lower_inner_product = cls.strictly_lower_inner_product(
            tangent_vec_a, tangent_vec_b
        )
        return diag_inner_product + strictly_lower_inner_product

    def exp(self, tangent_vec, base_point):
        """Compute the Cholesky exponential map.

        Compute the Riemannian exponential at point base_point
        of tangent vector tangent_vec wrt the Cholesky metric.
        This gives a lower triangular matrix with positive elements.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., n, n]
            Riemannian exponential.
        """
        sl_base_point = Matrices.to_strictly_lower_triangular(base_point)
        sl_tangent_vec = Matrices.to_strictly_lower_triangular(tangent_vec)
        diag_base_point = Matrices.diagonal(base_point)
        diag_tangent_vec = Matrices.diagonal(tangent_vec)
        diag_product_expm = gs.exp(gs.divide(diag_tangent_vec, diag_base_point))

        sl_exp = sl_base_point + sl_tangent_vec
        diag_exp = gs.vec_to_diag(diag_base_point * diag_product_expm)
        return sl_exp + diag_exp

    def log(self, point, base_point):
        """Compute the Cholesky logarithm map.

        Compute the Riemannian logarithm at point base_point,
        of point wrt the Cholesky metric.
        This gives a tangent vector at point base_point.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        log : array-like, shape=[..., n, n]
            Riemannian logarithm.
        """
        sl_base_point = Matrices.to_strictly_lower_triangular(base_point)
        sl_point = Matrices.to_strictly_lower_triangular(point)
        diag_base_point = Matrices.diagonal(base_point)
        diag_point = Matrices.diagonal(point)
        diag_product_logm = gs.log(gs.divide(diag_point, diag_base_point))

        sl_log = sl_point - sl_base_point
        diag_log = gs.vec_to_diag(diag_base_point * diag_product_logm)
        return sl_log + diag_log

    def squared_dist(self, point_a, point_b):
        """Compute the Cholesky Metric squared distance.

        Compute the Riemannian squared distance between point_a and point_b.

        Parameters
        ----------
        point_a : array-like, shape=[..., n, n]
            Point.
        point_b : array-like, shape=[..., n, n]
            Point.

        Returns
        -------
        _ : array-like, shape=[...]
            Riemannian squared distance.
        """
        log_diag_a = gs.log(Matrices.diagonal(point_a))
        log_diag_b = gs.log(Matrices.diagonal(point_b))
        diag_diff = log_diag_a - log_diag_b
        squared_dist_diag = gs.sum((diag_diff) ** 2, axis=-1)

        sl_a = Matrices.to_strictly_lower_triangular(point_a)
        sl_b = Matrices.to_strictly_lower_triangular(point_b)
        sl_diff = sl_a - sl_b
        squared_dist_sl = Matrices.frobenius_product(sl_diff, sl_diff)
        return squared_dist_sl + squared_dist_diag


class InvariantPositiveLowerTriangularMatricesMetric(_InvariantMetricMatrix):
    """Invariant metric on the positive lower triangular matrices."""

    def inner_product_at_identity(self, tangent_vec_a, tangent_vec_b):
        """Compute inner product at tangent space at identity.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n, n]
            First tangent vector at identity.
        tangent_vec_b : array-like, shape=[..., n, n]
            Second tangent vector at identity.

        Returns
        -------
        inner_prod : array-like, shape=[...]
            Inner-product of the two tangent vectors.
        """
        if Matrices.is_diagonal(self.metric_mat_at_identity):
            return super().inner_product_at_identity(tangent_vec_a, tangent_vec_b)

        return gs.dot(
            gs.tril_to_vec(tangent_vec_a),
            gs.matvec(self.metric_mat_at_identity, gs.tril_to_vec(tangent_vec_b)),
        )


class UnitNormedRowsPLTDiffeo(Diffeo):
    """A diffeomorphism from UnitNormedRowsPLTDMatrices to OpenHemispheresProduct.

    A diffeomorphism between the space of lower triangular matrices with
    positive diagonal and unit normed rows and the product space of
    successively increasing factor-dimension open hemispheres.

    Given the way `OpenHemisphere` is implemented, i.e. the first component
    is positive, rows need to be flipped.
    """

    def __init__(self, n):
        super().__init__()
        self.n = n

    def __call__(self, base_point):
        """Diffeomorphism at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        image_point : array-like, shape=[..., space_dim]
            Image point.
        """
        return gs.concatenate(
            [gs.flip(base_point[..., i, : i + 1], axis=-1) for i in range(1, self.n)],
            axis=-1,
        )

    def _from_product_to_mat(self, vec, ones=True):
        r"""Transform vector into matrix.

        A generalization of inverse diffeomorphism and inverse tangent
        diffeomorphism, where we can choose if the first element is one
        (diffeomorphism) or zero (tangent diffeormorphism).

        Parameters
        ----------
        vec : array-like, shape=[..., space_dim]

        Returns
        -------
        mat : array-like, shape=[..., n, n]
        """
        n = self.n
        flipped_vec = gs.flip(vec, axis=-1)
        batch_shape = vec.shape[:-1]

        gen_first = gs.ones if ones else gs.zeros

        first = gen_first(batch_shape + (1,))

        flipped_vec_ = gs.concatenate([flipped_vec, first], axis=-1)

        indices = sorted(zip(*gs.tril_indices(n)), key=lambda x: x[0], reverse=True)

        for n_batch in reversed(batch_shape):
            indices = [(k,) + indices_ for k in range(n_batch) for indices_ in indices]

        return gs.array_from_sparse(
            indices, gs.flatten(flipped_vec_), batch_shape + (n, n)
        )

    def inverse(self, image_point):
        r"""Inverse diffeomorphism at base point.

        :math:`f^{-1}: N \rightarrow M`

        Parameters
        ----------
        image_point : array-like, shape=[..., space_dim]
            Image point.

        Returns
        -------
        base_point : array-like, shape=[..., n, n]
            Base point.
        """
        return self._from_product_to_mat(image_point)

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        df_p is a linear map from T_pM to T_f(p)N.
        The diffeomorphism is linear, so its pushforward is itself

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.
        image_point : array-like, shape=[..., space_dim]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., space_dim]
            Image tangent vector at image of the base point.
        """
        return self(tangent_vec)

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        df^-1_p is a linear map from T_f(p)N to T_pM

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., space_dim]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., space_dim]
            Image point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *]
            Tangent vector at base point.
        """
        return self._from_product_to_mat(image_tangent_vec, ones=False)


class UnitNormedRowsPLTMatrices(DiffeomorphicManifold):
    """Space of positive lower triangular matrices with unit-normed rows.

    Rows are unit normed wrt Euclidean norm.

    For more details, check section 7.4.1 of [T2022]_.
    """

    def __init__(self, n, equip=True):
        diffeo = UnitNormedRowsPLTDiffeo(n=n)

        image_space = OpenHemispheresProduct(n=n) if n > 2 else OpenHemisphere(dim=1)

        super().__init__(
            dim=int(n * (n + 1) / 2),
            shape=(n, n),
            diffeo=diffeo,
            image_space=image_space,
            intrinsic=False,
            equip=equip,
        )
        self.n = n
        self.embedding_space = PositiveLowerTriangularMatrices(n=n, equip=False)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return UnitNormedRowsPLTMatricesPullbackMetric

    def belongs(self, point, atol=gs.atol):
        """Check if a point belongs to the unit normed plt matrices."""
        is_plt = self.embedding_space.belongs(point)
        is_unit_normed = self.image_space.belongs(self.diffeo(point))
        return gs.logical_and(is_plt, is_unit_normed)

    def is_tangent(self, vector, base_point, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector: array-like, shape = [..., *point_shape]
            Vector.
        base_point: array-like, shape = [..., *point_shape]
            Point on the manifold.
        atol: float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent: bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        is_lt = self.embedding_space.is_tangent(vector, base_point)
        first_is_zero = gs.isclose(vector[..., 0, 0], 0.0)

        image_point = self.diffeo(base_point)
        image_vector = self.diffeo.tangent(
            vector, base_point=base_point, image_point=image_point
        )
        image_is_tangent = self.image_space.is_tangent(
            image_vector, image_point, atol=atol
        )

        return gs.logical_and(
            gs.logical_and(is_lt, first_is_zero),
            image_is_tangent,
        )


class UnitNormedRowsPLTMatricesPullbackMetric(PullbackDiffeoMetric):
    """Pullback diffeo metric on `UnitNormedRowsPLTMatrices`.

    Results from the pullback of a metric on the `OpenHemispheresProduct`.
    """

    def __init__(self, space):
        super().__init__(
            space=space, diffeo=space.diffeo, image_space=space.image_space
        )


class PLTUnitDiagMatrices(LevelSet):
    """Lower triangular matrices with unit diagonal.

    Parameters
    ----------
    n : int
        Integer representing the shapes of the matrices: n x n.
    equip : bool
        If True, equip space with default metric.
    """

    def __init__(self, n, equip=True):
        self.n = n
        super().__init__(
            dim=int(n * (n - 1) / 2),
            equip=equip,
        )

    def _define_embedding_space(self):
        """Define embedding space of the manifold.

        Returns
        -------
        embedding_space : Manifold
            Instance of Manifold.
        """
        return LowerTriangularMatrices(self.n, equip=False)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return MatricesMetric

    def submersion(self, point):
        """Submersion that defines the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]

        Returns
        -------
        submersed_point : array-like, shape=[..., n]
        """
        return Matrices.diagonal(point) - gs.ones(self.n)

    def tangent_submersion(self, vector, point):
        """Tangent submersion.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
        point : array-like, shape=[..., n, n]

        Returns
        -------
        submersed_vector : array-like, shape=[..., n]
        """
        submersed_vector = Matrices.diagonal(vector)
        if point is not None and point.ndim > vector.ndim:
            return gs.broadcast_to(submersed_vector, point.shape[:-1])

        return submersed_vector

    def projection(self, point):
        """Project a point to the manifold.

        Parameters
        ----------
        point: array-like, shape[..., n, n]
            Point to project.

        Returns
        -------
        proj_point: array-like, shape[..., n, n]
            Projected point.
        """
        proj_point = self.embedding_space.projection(point)
        return proj_point + gs.vec_to_diag(
            1 - gs.diagonal(proj_point, axis1=-2, axis2=-1)
        )

    def to_tangent(self, vector, base_point):
        r"""Project a matrix to the tangent space at a base point.

        The tangent space of :math:`\mathrm{LT}^1(n)` is the space of
        strictly lower triangular matrices, :math:`\mathrm{LT}^0`.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
            Vector.
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        """
        tangent_vec = self.embedding_space.to_tangent(vector, base_point)
        return tangent_vec - gs.vec_to_diag(
            gs.diagonal(tangent_vec, axis1=-2, axis2=-1)
        )

    def random_point(self, n_samples=1, bound=1.0):
        """Sample LT^1 matrices by projecting random LT matrices.

        Parameters
        ----------
        n_samples : int
            Number of samples.
        bound : float
            Bound of the interval in which to sample.
            Optional, default: 1.

        Returns
        -------
        cor : array-like, shape=[n_samples, n, n]
            Sample of full-rank correlation matrices.
        """
        plt = self.embedding_space.random_point(n_samples, bound=bound)
        return self.projection(plt)


class LowerMatrixLog(Diffeo):
    """Matrix logarithm diffeomorphism.

    A diffeomorphism between the lower triangular matrices
    with unit diagonal and strictly lower triangular matrices.
    """

    @staticmethod
    def __call__(base_point):
        r"""Compute the matrix log.

        .. math::

            \log(Z) = \sum_{k=1}^{n-1} \frac{(-1)^{k+1}}{k}
            \left(Z-I_n\right)^k


        Parameters
        ----------
        base_point : array_like, shape=[..., n, n]
            Symmetric matrix.

        Returns
        -------
        log : array_like, shape=[..., n, n]
            Matrix logarithm of base_point.
        """
        n = base_point.shape[-1]

        result = gs.zeros_like(base_point)
        aux = base_point - gs.eye(n)
        for k in range(1, n):
            coeff = (-1) ** (k + 1) / k
            # aux = aux @ aux could be used, but accumulates too much error
            result += coeff * gs.linalg.matrix_power(aux, k)

        return result

    @staticmethod
    def inverse(image_point):
        r"""Compute the matrix exponential.

        .. math::

            \exp(L) = \sum_{k=0}^{n-1} \frac{1}{k!} L^k

        Parameters
        ----------
        image_point : array_like, shape=[..., n, n]
            Symmetric matrix.

        Returns
        -------
        exponential : array_like, shape=[..., n, n]
            Exponential of image_point.
        """
        n = image_point.shape[-1]

        result = gs.zeros_like(image_point) + gs.eye(n)
        factorial = 1
        for k in range(1, n):
            factorial *= k
            # aux = aux @ aux could be used, but accumulates too much error
            result += 1 / factorial * gs.linalg.matrix_power(image_point, k)

        return result

    @classmethod
    def tangent(cls, tangent_vec, base_point=None, image_point=None):
        r"""Tangent diffeomorphism at base point.

        .. math::

            d_z \log (V) = \sum_{k=1}^{n-1}\left(\frac{(-1)^{k+1}}{k}
            \sum_{j=1}^k\left((Z-I)^{j-1} V(Z-I)^{k-j}\right)\right)

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n, n]
            Base point.
        image_point : array-like, shape=[..., n, n]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., n, n]
            Image tangent vector at image of the base point.
        """

        def _matrix_power(power):
            return gs.linalg.matrix_power(
                aux,
                power,
            )

        if base_point is None:
            base_point = cls.inverse(image_point)

        n = tangent_vec.shape[-1]

        result = gs.zeros_like(tangent_vec)
        aux = base_point - gs.eye(n)
        for k in range(1, n):
            comp1 = (-1) ** (k + 1) / k
            comp2 = gs.zeros_like(tangent_vec)
            for j in range(1, k + 1):
                term = _matrix_power(j - 1) @ tangent_vec @ _matrix_power(k - j)
                comp2 += term
            result += comp1 * comp2
        return result

    @classmethod
    def inverse_tangent(cls, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse tangent diffeomorphism at image point.

        .. math::

            d_L \exp(X) = \sum_{k=1}^{n-1}\left(\frac{1}{k!}
            \sum_{j=1}^k\left(L^{j-1} X L^{k-j}\right)\right)

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., n, n]
            Image tangent vector at image point.
        image_point : array-like, shape=[..., n, n]
            Image point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        """

        def _matrix_power(power):
            return gs.linalg.matrix_power(image_point, power)

        if image_point is None:
            image_point = cls.__call__(base_point)

        n = image_tangent_vec.shape[-1]

        result = gs.zeros_like(image_tangent_vec)
        factorial = 1
        for k in range(1, n):
            factorial *= k
            comp1 = 1 / factorial
            comp2 = gs.zeros_like(image_tangent_vec)
            for j in range(1, k + 1):
                term = _matrix_power(j - 1) @ image_tangent_vec @ _matrix_power(k - j)
                comp2 += term
            result += comp1 * comp2
        return result



from geomstats.test_cases.geometry.lie_algebra import MatrixLieAlgebraTestCase


class SkewSymmetricMatricesTestCase(MatrixLieAlgebraTestCase):
    def test_baker_campbell_hausdorff_with_basis(self, atol):
        # valid for n > 2
        if self.space.n < 3:
            return

        fb = self.space.basis[0]
        sb = self.space.basis[1]

        fb_sb_bracket = self.space.bracket(fb, sb)
        expected1 = fb + sb
        expected2 = expected1 + 0.5 * fb_sb_bracket
        expected3 = (
            expected2
            + 1.0 / 12.0 * self.space.bracket(fb, fb_sb_bracket)
            - 1.0 / 12.0 * self.space.bracket(sb, fb_sb_bracket)
        )
        expected4 = expected3 - 1.0 / 24.0 * self.space.bracket(
            sb, self.space.bracket(fb, fb_sb_bracket)
        )
        expected = [expected1, expected2, expected3, expected4]
        for order in range(1, 5):
            self.test_baker_campbell_hausdorff(
                matrix_a=fb,
                matrix_b=sb,
                order=order,
                expected=expected[order - 1],
                atol=atol,
            )



"""Boundary value problem solvers implementation."""

import scipy

import geomstats.backend as gs
from geomstats.numerics._common import result_to_backend_type


class ScipySolveBVP:
    """Wrapper for scipy.integrate.solve_bvp."""

    def __init__(self, tol=1e-3, max_nodes=1000, bc_tol=None, save_result=False):
        self.tol = tol
        self.max_nodes = max_nodes
        self.bc_tol = bc_tol

        self.save_result = save_result
        self.result_ = None

    def integrate(self, fun, bc, x, y, fun_jac=None, bc_jac=None):
        """Solve a boundary value problem for a system of ODEs."""

        def fun_(t, state):
            return fun(t, gs.from_numpy(state))

        def bc_(state_0, state_1):
            return bc(gs.from_numpy(state_0), gs.from_numpy(state_1))

        if fun_jac is not None:

            def fun_jac_(t, state):
                return fun_jac(t, gs.from_numpy(state))

        else:
            fun_jac_ = None

        result = scipy.integrate.solve_bvp(
            fun_,
            bc_,
            x,
            y,
            tol=self.tol,
            max_nodes=self.max_nodes,
            fun_jac=fun_jac_,
            bc_jac=bc_jac,
            bc_tol=self.bc_tol,
        )

        result = result_to_backend_type(result)
        if self.save_result:
            self.result_ = result

        return result



"""Geometric median."""

import logging

from sklearn.base import BaseEstimator

import geomstats.backend as gs


class GeometricMedian(BaseEstimator):
    """Geometric median.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    max_iter : int
        Maximum number of iterations for the algorithm.
        Optional, default : 100
    lr : float
        Learning rate to be used for the algorithm.
        Optional, default : 1.0
    init_point : array-like, shape=[*space.shape]
        Initialization to be used in the start.
        Optional, default : None, in which case it uses last sample.
    print_every : int
        Print updated median after print_every iterations.
        Optional, default : None
    epsilon : float
        Tolerance for stopping the algorithm (distance between two successive
        estimates).
        Optional, default : gs.atol

    Attributes
    ----------
    estimate_ : array-like, shape=[*space.shape]
        If fit, geometric median.

    Notes
    -----
    * Required metric methods: `dist`, `log`, `exp`.

    References
    ----------
    .. [FVJ2009] Fletcher PT, Venkatasubramanian S and Joshi S.
        "The geometric median on Riemannian manifolds with application to
        robust atlas estimation", NeuroImage, 2009
        https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2735114/
    """

    def __init__(
        self,
        space,
        max_iter=100,
        lr=1.0,
        init_point=None,
        print_every=None,
        epsilon=gs.atol,
    ):
        self.space = space
        self.max_iter = max_iter
        self.lr = lr
        self.init_point = init_point
        self.print_every = print_every
        self.epsilon = epsilon

        self.estimate_ = None

    def _iterate_once(self, current_median, X, weights, lr):
        """Compute a single iteration of Weiszfeld algorithm.

        Parameters
        ----------
        current_median : array-like, shape=[*space.shape]
            Current median.
        X : array-like, shape=[n_samples, *space.shape]
            Training input samples.
        weights : array-like, shape=[n_samples,]
            Weights for weighted sum.
        lr : float
            Learning rate for the current iteration.

        Returns
        -------
        updated_median : array-like, shape=[*metric.shape]
            Updated median after single iteration.
        """
        dists = self.space.metric.dist(current_median, X)
        is_non_zero = dists > gs.atol
        if not gs.any(is_non_zero):
            return current_median

        w = weights[is_non_zero] / dists[is_non_zero]
        logs = self.space.metric.log(X[is_non_zero], current_median)
        v_k = gs.einsum("n,n...->...", w / gs.sum(w), logs)
        updated_median = self.space.metric.exp(lr * v_k, current_median)
        return updated_median

    def fit(self, X, y=None, weights=None):
        """Compute the weighted geometric median.

        Compute the geometric median on manifold using Weiszfeld algorithm.

        Parameters
        ----------
        X : array-like, shape=[n_samples, *metric.shape]
            Training input samples.
        y : None
            Target values. Ignored.
        weights : array-like, shape=[n_samples,]
            Weights associated to the samples.
            Optional, default: None, in which case it is equally weighted.

        Returns
        -------
        self : object
            Returns self.
        """
        median = X[-1] if self.init_point is None else self.init_point
        if weights is None:
            weights = gs.ones(X.shape[0])
        weights /= gs.sum(weights)

        for iteration in range(self.max_iter):
            new_median = self._iterate_once(median, X, weights, self.lr)
            shift = self.space.metric.dist(new_median, median)

            if shift < self.epsilon:
                break
            median = new_median

            if self.print_every and (iteration + 1) % self.print_every == 0:
                logging.info(f"median at iteration {iteration+1}:\n{median}")
        else:
            logging.warning(
                "Maximum number of iterations %s reached. "
                "The median may be inaccurate",
                self.max_iter,
            )

        self.estimate_ = median
        return self



import pytest

import geomstats.backend as gs
from geomstats.geometry.matrices import Matrices
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.base import LevelSetTestCase


class FullRankCorrelationMatricesTestCase(LevelSetTestCase):
    def test_from_covariance(self, point, expected, atol):
        point_ = self.space.from_covariance(point)
        self.assertAllClose(point_, expected, atol=atol)

    @pytest.mark.random
    def test_from_covariance_belongs(self, n_points, atol):
        spd_point = self.space.embedding_space.random_point(n_points)

        point = self.space.from_covariance(spd_point)
        expected = gs.ones(n_points, dtype=bool)
        self.test_belongs(point, expected, atol)

    @pytest.mark.vec
    def test_from_covariance_vec(self, n_reps, atol):
        point = self.space.embedding_space.random_point()
        expected = self.space.from_covariance(point)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_diag_action(self, diagonal_vec, point, expected, atol):
        res = self.space.diag_action(diagonal_vec, point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_diag_action_vec(self, n_reps, atol):
        point = self.space.embedding_space.random_point()
        diagonal_vec = Matrices.diagonal(point)

        expected = self.space.diag_action(diagonal_vec, point)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    diagonal_vec=diagonal_vec, point=point, expected=expected, atol=atol
                )
            ],
            arg_names=["diagonal_vec", "point"],
            expected_name="expected",
            n_reps=n_reps,
            vectorization_type="sym",
        )
        self._test_vectorization(vec_data)



"""Module exposing the GeneralLinear group class."""

import geomstats.algebra_utils as utils
import geomstats.backend as gs
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.lie_algebra import MatrixLieAlgebra
from geomstats.geometry.lie_group import MatrixLieGroup
from geomstats.geometry.matrices import Matrices, MatricesMetric


class GeneralLinear(MatrixLieGroup, VectorSpaceOpenSet):
    """Class for the general linear group GL(n) and its identity component.

    If `positive_det=True`, this is the connected component of the identity,
    i.e. the space of matrices with positive determinant.

    Parameters
    ----------
    n : int
        Integer representing the shape of the matrices: n x n.
    positive_det : bool
        Whether to restrict to the identity connected component of the
        general linear group, i.e. matrices with positive determinant.
        Optional, default: False.
    """

    def __init__(self, n, positive_det=False, equip=True):
        self.n = n
        super().__init__(
            dim=n**2,
            embedding_space=Matrices(n, n, equip=False),
            representation_dim=n,
            lie_algebra=SquareMatrices(n, equip=False),
            equip=equip,
        )

        self.positive_det = positive_det

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return MatricesMetric

    def projection(self, point):
        r"""Project a matrix to the general linear group.

        As GL(n) is dense in the space of matrices, this is not a projection
        per se, but a regularization if the matrix is not already invertible:
        :math:`X + \epsilon I_n` is returned where :math:`\epsilon=gs.atol`
        is returned for an input X.

        Parameters
        ----------
        point : array-like, shape=[..., dim_embedding]
            Point in embedding manifold.

        Returns
        -------
        projected : array-like, shape=[..., dim_embedding]
            Projected point.
        """
        belongs = self.belongs(point)
        regularization = gs.einsum(
            "...,ij->...ij", gs.where(~belongs, gs.atol, 0.0), self.identity
        )
        projected = point + regularization
        if self.positive_det:
            det = gs.linalg.det(point)
            return utils.flip_determinant(projected, det)
        return projected

    def belongs(self, point, atol=gs.atol):
        """Check if a matrix is invertible and of the right shape.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Matrix to be checked.
        atol : float
            Tolerance threshold for the determinant.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean denoting if point is in GL(n).
        """
        has_right_size = self.embedding_space.belongs(point)
        if gs.all(has_right_size):
            det = gs.linalg.det(point)
            return det > atol if self.positive_det else gs.abs(det) > atol
        return has_right_size

    def random_point(self, n_samples=1, bound=1.0, n_iter=100):
        """Sample in GL(n) from the normal distribution.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound: float
            This parameter is ignored
        n_iter : int
            Maximum number of trials to sample a matrix with positive det.
            Optional, default: 100.

        Returns
        -------
        samples : array-like, shape=[..., n, n]
            Point sampled on GL(n).
        """
        n = self.n
        sample = []
        n_accepted, iteration = 0, 0
        criterion_func = (lambda x: x) if self.positive_det else gs.abs
        while n_accepted < n_samples and iteration < n_iter:
            raw_samples = gs.random.normal(size=(n_samples - n_accepted, n, n))
            dets = gs.linalg.det(raw_samples)
            criterion = criterion_func(dets) > gs.atol
            if gs.any(criterion):
                sample.append(raw_samples[criterion])
                n_accepted += gs.sum(criterion)
            iteration += 1
        if n_samples == 1:
            return sample[0][0]
        return gs.concatenate(sample)

    @classmethod
    def orbit(cls, point, base_point=None):
        r"""
        Compute the one-parameter orbit of base_point passing through point.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Target point.
        base_point : array-like, shape=[..., n, n], optional
            Base point.
            Optional, defaults to identity if None.

        Returns
        -------
        path : callable
            One-parameter orbit.
            Satisfies `path(0) = base_point` and `path(1) = point`.

        Notes
        -----
        Denoting `point` by :math:`g` and `base_point` by :math:`h`,
        the orbit :math:`\gamma` satisfies:

        .. math::

            \gamma(t) = {\mathrm e}^{t X} \cdot h \\
            \quad \text{with} \quad\\
            {\mathrm e}^{X} = g h^{-1}

        The path is not uniquely defined and depends on the choice of :math:`V`
        returned by :py:meth:`log`.

        Vectorization
        -------------
        Return a collection of trajectories (4-D array)
        from a collection of input matrices (3-D array).
        """
        tangent_vec = cls.log(point, base_point)

        if base_point is not None and gs.ndim(base_point) < gs.ndim(tangent_vec):
            base_point = gs.broadcast_to(base_point, tangent_vec.shape)
        is_vec = gs.ndim(tangent_vec) > 2

        def path(time):
            vecs = gs.einsum("t,...ij->...tij", time, tangent_vec)
            if is_vec and base_point is None:
                return gs.stack([cls.exp(vecs_) for vecs_ in vecs])

            if is_vec:
                return gs.stack(
                    [
                        cls.exp(vecs_, base_point_)
                        for vecs_, base_point_ in zip(vecs, base_point)
                    ]
                )

            return cls.exp(vecs, base_point)

        return path


class SquareMatrices(MatrixLieAlgebra):
    """Lie algebra of the general linear group.

    This is the space of matrices.

    Parameters
    ----------
    n : int
        Integer representing the shape of the matrices: n x n.
    """

    def __init__(self, n, equip=True):
        self.n = n
        super().__init__(dim=n**2, representation_dim=n, equip=equip)
        self._mat_space = Matrices(n, n, equip=False)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return MatricesMetric

    def _create_basis(self):
        """Create the canonical basis of the space of matrices."""
        return self._mat_space.basis

    def basis_representation(self, matrix_representation):
        """Compute the coefficient in the usual matrix basis.

        This simply flattens the input.

        Parameters
        ----------
        matrix_representation : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        basis_representation : array-like, shape=[..., dim]
            Representation in the basis.
        """
        return self._mat_space.flatten(matrix_representation)

    def matrix_representation(self, basis_representation):
        """Compute the matrix representation for the given basis coefficients.

        This simply reshapes the input into a square matrix.

        Parameters
        ----------
        basis_representation : array-like, shape=[..., dim]
            Coefficients in the basis.

        Returns
        -------
        matrix_representation : array-like, shape=[..., n, n]
            Matrix.
        """
        return self._mat_space.reshape(basis_representation)



"""Machinery to handle global control of dtypes.

Notes
-----
Functions starting with "_pre" are shared functions that just need access to
specific backend functions. e.g. `_pre_set_default_dtype` requires access to
`as_dtype`. `set_default_dtype` can then be created in each backend by doing
`set_default_dtype = _pre_set_default_dtype(as_dtype)`. The same principle
applies to ("_pre") decorators. This decreases code duplication, while being
able to avoid (dirty) circular imports.
"""

import functools
import inspect
import types

from geomstats._backend import _backend_config as _config

_TO_UPDATE_FUNCS_DTYPE = []
_TO_UPDATE_FUNCS_KW_DTYPE = []


_MAP_FLOAT_TO_COMPLEX = {
    "float32": "complex64",
    "float64": "complex128",
}


def _copy_func(func):
    """Copy function."""
    new_func = types.FunctionType(
        func.__code__,
        func.__globals__,
        func.__name__,
        func.__defaults__,
        func.__closure__,
    )
    new_func.__dict__.update(func.__dict__)
    new_func.__kwdefaults__ = func.__kwdefaults__

    return new_func


def _get_dtype_pos_in_defaults(func):
    """Get dtype position in defaults."""
    pos = 0
    for name, parameter in inspect.signature(func).parameters.items():
        if name == "dtype":
            return pos
        if parameter.default is not inspect._empty:
            pos += 1

    raise Exception("dtype is not kwarg")


def _update_default_dtypes():
    """Update default dtype of functions.

    How it works?
    -------------
    "dtype" is updated in __defaults__ or __kwdefaults__ to default dtype.

    Every time default dtype is changed, all the functions that follow this
    strategy will have their default dtype updated.

    It (mutably) changes function defaults. For external functions, copy the
    function first to avoid surprising users.
    """
    for func in _TO_UPDATE_FUNCS_DTYPE:
        pos = _get_dtype_pos_in_defaults(func)
        defaults = list(func.__defaults__)
        defaults[pos] = _config.DEFAULT_DTYPE
        func.__defaults__ = tuple(defaults)

    for func in _TO_UPDATE_FUNCS_KW_DTYPE:
        func.__kwdefaults__["dtype"] = _config.DEFAULT_DTYPE


def _modify_func_default_dtype(copy=True, kw_only=False, target=None):
    """Modify function default dtype by acting directly in the target object.

    Parameters
    ----------
    copy: bool
        If true, copies function before changing dtype.
    kw_only : bool
        If true, it is assumed dtype is kwarg only argument.

    How it works?
    -------------
    This decorator only collects functions. Default dtype is modified only
    when default dtype is changed (see `_update_default_dtypes`).
    """

    def _decorator(func):
        new_func = _copy_func(func) if copy else func

        if kw_only:
            _TO_UPDATE_FUNCS_KW_DTYPE.append(new_func)
        else:
            _TO_UPDATE_FUNCS_DTYPE.append(new_func)

        return new_func

    if target is None:
        return _decorator

    return _decorator(target)


def get_default_dtype():
    """Get backend default float dtype."""
    return _config.DEFAULT_DTYPE


def get_default_cdtype():
    """Get backend default complex dtype."""
    return _config.DEFAULT_COMPLEX_DTYPE


def _dyn_update_dtype(dtype_pos=None, target=None):
    """Update (dynamically) function dtype.

    Parameters
    ----------
    dtype_pos : int
        Position of "dtype" argument.

    How it works?
    -------------
    When the function is called, it verifies if dtype is passed. If not, it
    uses default dtype.
    """

    def _decorator(func):
        @functools.wraps(func)
        def _wrapped(*args, **kwargs):
            if dtype_pos is not None and len(args) > dtype_pos:
                args = list(args)
                args[dtype_pos] = _config.DEFAULT_DTYPE

            else:
                if kwargs.get("dtype") is None:
                    kwargs["dtype"] = _config.DEFAULT_DTYPE

            return func(*args, **kwargs)

        return _wrapped

    if target is None:
        return _decorator

    return _decorator(target)


def _pre_set_default_dtype(as_dtype):
    def set_default_dtype(value):
        """Set backend default dtype.

        Parameters
        ----------
        value : str
            Possible values are "float32" as "float64".
        """
        _config.DEFAULT_DTYPE = as_dtype(value)
        _config.DEFAULT_COMPLEX_DTYPE = as_dtype(_MAP_FLOAT_TO_COMPLEX[value])

        _update_default_dtypes()

        return get_default_dtype()

    return set_default_dtype


def _pre_cast_out_from_dtype(cast, is_floating, is_complex):
    def _cast_out_from_dtype(dtype_pos=None, target=None):
        """Cast output based on default dtype.

        Useful to wrap functions which output dtype cannot be (fully) controlled
        or for which is useful to run it without controlling dtype and cast
        afterwards (e.g. array).

        Parameters
        ----------
        dtype_pos : int
            Position of "dtype" argument.

        How it works?
        -------------
        Function is called normally. If output is float or complex, then it
        checks if is of expected dtype. If not, cast is performed.
        """

        def _decorator(func):
            @functools.wraps(func)
            def _wrapped(*args, **kwargs):
                out = func(*args, **kwargs)

                if is_floating(out) or is_complex(out):
                    if dtype_pos is not None and len(args) > dtype_pos:
                        dtype = args[dtype_pos]
                    else:
                        dtype = kwargs.get(
                            "dtype",
                            (
                                _config.DEFAULT_DTYPE
                                if is_floating(out)
                                else _config.DEFAULT_COMPLEX_DTYPE
                            ),
                        )

                    if out.dtype != dtype:
                        return cast(out, dtype)

                return out

            return _wrapped

        if target is None:
            return _decorator

        return _decorator(target)

    return _cast_out_from_dtype


def _pre_add_default_dtype_by_casting(cast):
    def _add_default_dtype_by_casting(target=None):
        """Add default dtype as function argument.

        Behavior is achieved by casting output (not ideal, but impoosible to
        avoid without acting directly in the backends themselves).

        How it works?
        -------------
        Function is called normally. If output is float or complex, then it
        checks if is of expected dtype. If not, cast is performed.

        The difference to `_cast_out_from_dtype` is that wrapped functions do
        not accept dtype.
        """

        def _decorator(func):
            @functools.wraps(func)
            def _wrapped(*args, dtype=None, **kwargs):
                if dtype is None:
                    dtype = _config.DEFAULT_DTYPE

                out = func(*args, **kwargs)
                if out.dtype != dtype:
                    return cast(out, dtype)
                return out

            return _wrapped

        if target is None:
            return _decorator

        return _decorator(target)

    return _add_default_dtype_by_casting


def _pre_cast_fout_to_input_dtype(cast, is_floating):
    def _cast_fout_to_input_dtype(target=None):
        """Cast out func if float and not accordingly to input.

        It is required e.g. for scipy when result is innacurate.

        How it works?
        -------------
        Function is called normally. If output is float, then it
        checks if is of expected dtype (input dtype). If not, cast is performed.

        The difference to `_cast_out_from_dtype` is that output is expected to
        have same type as input.
        """

        def _decorator(func):
            @functools.wraps(func)
            def _wrapped(x, *args, **kwargs):
                out = func(x, *args, **kwargs)
                if is_floating(out) and out.dtype != x.dtype:
                    return cast(out, x.dtype)
                return out

            return _wrapped

        if target is None:
            return _decorator

        return _decorator(target)

    return _cast_fout_to_input_dtype


def _pre_cast_out_to_input_dtype(cast, is_floating, is_complex, as_dtype, dtype_as_str):
    def _cast_out_to_input_dtype(target=None):
        """Cast out func if float or complex and not accordingly to input.

        How it works?
        -------------
        Function is called normally.
        If output is float, then it checks if is of expected dtype
        (input dtype). If not, cast is performed.
        If output is complex, then if first check if input is complex, if not
        it verifies the required precision for complex dtype and casts
        accordingly (if necessary)
        """

        def _decorator(func):
            @functools.wraps(func)
            def _wrapped(x, *args, **kwargs):
                out = func(x, *args, **kwargs)

                if is_floating(out):
                    if out.dtype != x.dtype:
                        return cast(out, x.dtype)
                elif is_complex(out):
                    if is_complex(x):
                        cmp_dtype = x.dtype
                    else:
                        float_name = dtype_as_str(x.dtype)
                        cmp_dtype = as_dtype(f"complex{int((float_name[-2:]))*2}")

                    if out.dtype != cmp_dtype:
                        return cast(out, cmp_dtype)

                return out

            return _wrapped

        if target is None:
            return _decorator

        return _decorator(target)

    return _cast_out_to_input_dtype


def _pre_allow_complex_dtype(cast, complex_dtypes):
    def _allow_complex_dtype(target=None):
        """Allow complex type by calling the function twice.

        Assumes function do not support dtype.

        How it works?
        -------------
        Function is called twice if dtype is complex.
        Output is casted if not corresponding to expected dtype.
        """

        def _decorator(func):
            @functools.wraps(func)
            def _wrapped(*args, dtype=None, **kwargs):
                out = func(*args, **kwargs)
                if dtype in complex_dtypes:
                    out = out + 1j * func(*args, **kwargs)

                if out.dtype != dtype:
                    return cast(out, dtype)

                return out

            return _wrapped

        if target is None:
            return _decorator

        return _decorator(target)

    return _allow_complex_dtype


def _np_box_unary_scalar(target=None):
    """Update dtype if input is float in unary operations.

    How it works?
    -------------
    If dtype is float, then default dtype is passed as argument.
    """

    def _decorator(func):
        @functools.wraps(func)
        def _wrapped(x, *args, **kwargs):
            if isinstance(x, float):
                return func(x, *args, dtype=_config.DEFAULT_DTYPE, **kwargs)

            return func(x, *args, **kwargs)

        return _wrapped

    if target is None:
        return _decorator

    return _decorator(target)


def _np_box_binary_scalar(target=None):
    """Update dtype if input is float in binary operations.

    How it works?
    -------------
    If dtype is float, then default dtype is passed as argument.
    """

    def _decorator(func):
        @functools.wraps(func)
        def _wrapped(x1, x2, *args, **kwargs):
            if isinstance(x1, float):
                return func(x1, x2, *args, dtype=_config.DEFAULT_DTYPE, **kwargs)

            return func(x1, x2, *args, **kwargs)

        return _wrapped

    if target is None:
        return _decorator

    return _decorator(target)



"""Interpolation machinery."""

from abc import ABC, abstractmethod

import geomstats.backend as gs


class Interpolator(ABC):
    """Abstract class for interpolator."""

    def __call__(self, t):
        """Interpolate data.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.

        Returns
        -------
        point : array-like, shape=[..., n_time, *point_shape]
        """
        return self.interpolate(t)

    @abstractmethod
    def interpolate(self, t):
        """Interpolate data.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.

        Returns
        -------
        point : array-like, shape=[..., n_time, *point_shape]
        """


class _LinearInterpolator1D(Interpolator, ABC):
    def __init__(self, data, point_ndim=1):
        self.data = data
        self.point_ndim = point_ndim

        time_axis = -(point_ndim + 1)
        self._n_times = self.data.shape[time_axis]

    @abstractmethod
    def _from_t_to_interval(self, t):
        """Get interval index from time.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.

        Returns
        -------
        interval_index : array-like, shape=[n_times]
        """

    def _get_ratio(self, t, interval_index, end_index):
        """Get ratio within interval.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.
        interval_index : array-like, shape=[n_times]
        end_index : array-like, shape=[n_times]

        Returns
        -------
        ratio : array-like, shape=[n_time]
            Ratio of t within interval.
        """

    def interpolate(self, t):
        """Interpolate data.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.

        Returns
        -------
        point : array-like, shape=[..., n_time, *point_shape]
        """
        if not gs.is_array(t):
            t = gs.array([t])

        if gs.ndim(t) == 0:
            t = gs.expand_dims(t, axis=0)

        interval_index = self._from_t_to_interval(t)

        max_bound_reached = interval_index == self._n_times - 1

        end_index = gs.where(max_bound_reached, interval_index, interval_index + 1)

        point_ndim_slc = (slice(None),) * self.point_ndim
        initial_point = self.data[(..., interval_index) + point_ndim_slc]
        end_point = self.data[(..., end_index) + point_ndim_slc]

        ratio = self._get_ratio(t, interval_index, end_index)

        diff = end_point - initial_point
        ijk = "ijk"[: self.point_ndim]
        return initial_point + gs.einsum(f"t,...t{ijk}->...t{ijk}", ratio, diff)


class UniformUnitIntervalLinearInterpolator(_LinearInterpolator1D):
    """A 1D linear interpolator.

    Assumes interpolation occurs in the unit interval and
    data is uniformly sampled.

    Parameters
    ----------
    data : array-like, [..., *point_shape]
    point_ndim : int
        Dimension of point.
    """

    def __init__(self, data, point_ndim=1):
        super().__init__(data, point_ndim=point_ndim)
        self._delta = 1 / (self._n_times - 1)

    def _from_t_to_interval(self, t):
        """Get interval index from time.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.

        Returns
        -------
        interval_index : array-like, shape=[n_times]
        """
        return gs.cast(
            t // self._delta,
            dtype=gs.int32,
        )

    def _get_ratio(self, t, interval_index, end_index):
        """Get ratio within interval.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.
        interval_index : array-like, shape=[n_times]
            Ignored.
        end_index : array-like, shape=[n_times]
            Ignored.

        Returns
        -------
        ratio : array-like, shape=[n_time]
            Ratio of t within interval.
        """
        return gs.mod(t, self._delta) / self._delta


class LinearInterpolator1D(_LinearInterpolator1D):
    """A 1D linear interpolator.

    Assumes interpolation occurs in the unit interval.

    Parameters
    ----------
    times : array-like, [n_times]
        Times. Must be sorted.
    data : array-like, [..., *point_shape]
    point_ndim : int
        Dimension of point.
    """

    def __init__(self, times, data, point_ndim=1):
        super().__init__(data, point_ndim=point_ndim)
        self.times = times
        self._delta = self.times[1:] - self.times[:-1]

    def _from_t_to_interval(self, t):
        """Get interval index from time.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.

        Returns
        -------
        interval_index : array-like, shape=[n_times]
        """
        indices = gs.searchsorted(self.times, t) - 1
        return gs.where(indices < 0, 0, indices)

    def _get_ratio(self, t, interval_index, end_index):
        """Get ratio within interval.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.
        interval_index : array-like, shape=[n_times]
        end_index : array-like, shape=[n_times]

        Returns
        -------
        ratio : array-like, shape=[n_time]
            Ratio of t within interval.
        """
        delta = self._delta[interval_index]
        return (delta - (self.times[end_index] - t)) / delta



"""Scalar product of a Riemannian metric.

Define the product of a Riemannian metric with a scalar number.

Public Methods:
    register_scaled_method(func_name, scaling_type)

Public classes
    ScalarProductMetric

Lead author: John Harvey.
"""

from functools import wraps

import geomstats.backend as gs
import geomstats.errors


def register_scaled_method(func_name, scaling_type):
    """Register the scaling factor of a method of a RiemannianMetric.

    The ScalarProductMetric class rescales various methods of a
    RiemannianMetric by the correct factor. The default behaviour is to
    rescale linearly. This method allows the user to add a new method to be
    rescaled according to a different rule.

    Note that this method must be called before the ScalarProductMetric is
    instantiated. It does not affect objects which already exist.

    Parameters
    ----------
    func_name : str
        The name of a method from a RiemannianMetric object which must be
        rescaled.
    scaling_type : str, {'sqrt',
                         'linear',
                         'quadratic',
                         'inverse',
                         'inverse_sqrt'}
        How the method should be rescaled as a function of
        ScalarProductMetric.scale.

    """
    _ScaledMethodsRegistry._add_scaled_method(func_name, scaling_type)


def _wrap_attr(scaling_factor, func):
    @wraps(func)
    def response(*args, **kwargs):
        res = scaling_factor * func(*args, **kwargs)
        return res

    return response


class _ScaledMethodsRegistry:
    """Class to hold lists of methods and their scaling functions."""

    _SQRT_LIST = ["norm", "dist", "dist_broadcast", "dist_pairwise", "diameter"]
    _LINEAR_LIST = [
        "metric_matrix",
        "inner_product",
        "inner_product_derivative_matrix",
        "squared_norm",
        "squared_dist",
        "covariant_riemann_tensor",
    ]
    _QUADRATIC_LIST = []
    _INVERSE_LIST = [
        "cometric_matrix",
        "inner_coproduct",
        "hamiltonian",
        "sectional_curvature",
        "scalar_curvature",
    ]
    _INVERSE_SQRT_LIST = ["normalize", "random_unit_tangent_vec", "normal_basis"]
    _RESERVED_NAMES = ("underlying_metric", "scale")

    _SCALING_LISTS = [
        _SQRT_LIST,
        _LINEAR_LIST,
        _QUADRATIC_LIST,
        _INVERSE_LIST,
        _INVERSE_SQRT_LIST,
    ]
    _SCALING_NAMES = ["sqrt", "linear", "quadratic", "inverse", "inverse_sqrt"]

    @classmethod
    def _add_scaled_method(cls, func_name, scaling_type):
        """Configure ScalarProductMetric to scale an attribute.

        This method should be accessed via
        geomstats.geometry.scalar_product_metric.register_scaled_method
        """
        scaling_dict = dict(zip(cls._SCALING_NAMES, cls._SCALING_LISTS))

        for list_of_methods in cls._SCALING_LISTS:
            if func_name in list_of_methods:
                msg = (
                    f"'{func_name}' already has an assigned scaling rule "
                    "which cannot be changed."
                )
                raise ValueError(msg)
        if func_name in cls._RESERVED_NAMES:
            raise ValueError(f"'{func_name}' is reserved for internal use.")
        if func_name.startswith("_"):
            raise ValueError("Private methods cannot be rescaled")

        try:
            scaling_dict[scaling_type].append(func_name)
        except KeyError:
            msg = (
                f"'{scaling_type}' is not an admissible value. Please "
                "provide one of 'sqrt', 'linear', 'quadratic', "
                "'inverse', 'inverse_sqrt'."
            )
            raise ValueError(msg)

    @classmethod
    def _get_scaling_factor(cls, func_name, scale):
        if func_name in cls._SQRT_LIST:
            return gs.sqrt(scale)

        if func_name in cls._LINEAR_LIST:
            return scale

        if func_name in cls._QUADRATIC_LIST:
            return gs.power(scale, 2)

        if func_name in cls._INVERSE_LIST:
            return 1.0 / scale

        if func_name in cls._INVERSE_SQRT_LIST:
            return 1.0 / gs.sqrt(scale)
        return None


class ScalarProductMetric:
    """Class for scalar products of Riemannian and pseudo-Riemannian metrics.

    This class multiplies the (0,2) metric tensor 'space.metric' by a
    scalar 'scaling_factor'. Note that this does not scale distances by
    'scaling_factor'. That would require multiplication by the square of the
    scalar.

    The `space` is not automatically equipped with the `ScalarProductMetric`.

    An object of this type can also be instantiated by the expression
    scaling_factor * space.metric.

    This class acts as a wrapper for the underlying Riemannian metric. All
    public attributes apart from 'underlying_metric' and 'scaling_factor' are
    loaded from the underlying metric at initialization and rescaled by the
    appropriate factor. Changes to the underlying metric at runtime will not
    affect the attributes of this object.

    One exception to this is when the 'underlying_metric' is itself of type
    ScalarProductMetric. In this case, rather than wrapping the wrapper, the
    'underlying_metric' of the first ScalarProductMetric object is wrapped a
    second time with a new 'scaling_factor'.

    Parameters
    ----------
    space : Manifold or ComplexManifold
        A manifold equipped with a metric which is being scaled.
    scale : float
        The value by which to scale the metric. Note that this rescales the
        (0,2) metric tensor, so distances are rescaled by the square root of
        this.
    """

    def __init__(self, space, scale):
        """Load all attributes from the underlying metric."""
        geomstats.errors.check_positive(scale, "scale")
        if not hasattr(space, "metric"):
            raise TypeError("The variable 'space' must be equipped with a metric.")

        self._space = space

        if isinstance(space.metric, ScalarProductMetric):
            self.underlying_metric = space.metric.underlying_metric
            self.scale = scale * space.metric.scale
        else:
            self.underlying_metric = space.metric
            self.scale = scale

        for attr_name in dir(self.underlying_metric):
            if (
                attr_name.startswith("_")
                or attr_name in _ScaledMethodsRegistry._RESERVED_NAMES
            ):
                continue

            attr = getattr(self.underlying_metric, attr_name)
            if not callable(attr):
                try:
                    setattr(self, attr_name, attr)
                except AttributeError as ex:
                    if not isinstance(
                        getattr(type(self.underlying_metric), attr_name, None),
                        property,
                    ):
                        raise ex
            else:
                scale = _ScaledMethodsRegistry._get_scaling_factor(
                    attr_name, self.scale
                )
                method = attr if scale is None else _wrap_attr(scale, attr)
                setattr(self, attr_name, method)

    def __mul__(self, scalar):
        """Multiply the metric by a scalar.

        This method multiplies the (0,2) metric tensor by a scalar. Note that
        this does not scale distances by the scalar. That would require
        multiplication by the square of the scalar.

        Parameters
        ----------
        scalar : float
            The number by which to multiply the metric.

        Returns
        -------
        metric : ScalarProductMetric
            The metric multiplied by the scalar
        """
        if not isinstance(scalar, float):
            return NotImplemented
        if self != self._space.metric:
            raise ValueError(
                "A space must be equipped with this metric before it is scaled."
            )
        return ScalarProductMetric(self._space, scalar)

    def __rmul__(self, scalar):
        """Multiply the metric by a scalar.

        This method multiplies the (0,2) metric tensor by a scalar. Note that
        this does not scale distances by the scalar. That would require
        multiplication by the square of the scalar.

        Parameters
        ----------
        scalar : float
            The number by which to multiply the metric.

        Returns
        -------
        metric : ScalarProductMetric
            The metric multiplied by the scalar.
        """
        return self * scalar



"""Statistical Manifold of categorical distributions with the Fisher metric.

Lead author: Alice Le Brigant.
"""

from geomstats.information_geometry.multinomial import (
    MultinomialDistributions,
    MultinomialMetric,
)


class CategoricalDistributions(MultinomialDistributions):
    r"""Class for the manifold of categorical distributions.

    This is the set of `n+1`-tuples of positive reals that sum up to one,
    i.e. the `n`-simplex. Each point is the parameter of a categorical
    distribution, i.e. gives the probabilities of $n$ different outcomes
    in a single experiment.

    Attributes
    ----------
    dim : int
        Dimension of the manifold of categorical distributions. The
        number of outcomes is dim + 1.
    embedding_manifold : Manifold
        Embedding manifold.
    """

    def __init__(self, dim, equip=True):
        super().__init__(dim=dim, n_draws=1, equip=equip)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return CategoricalMetric


class CategoricalMetric(MultinomialMetric):
    """Class for the Fisher information metric on categorical distributions.

    The Fisher information metric on the $n$-simplex of categorical
    distributions parameters can be obtained as the pullback metric of the
    $n$-sphere using the componentwise square root.

    References
    ----------
    .. [K2003] R. E. Kass. The Geometry of Asymptotic Inference. Statistical
        Science, 4(3): 188 - 234, 1989.
    """



import geomstats.backend as gs
from geomstats.test.test_case import TestCase


class AgglomerativeHierarchicalClusteringTestCase(TestCase):
    def test_fit_two_clusters(self, estimator, dataset, expected):
        if gs.unique(expected).shape[0] > 2:
            raise ValueError("Expected only two clusters")

        estimator.fit(dataset)

        clustering_labels = estimator.labels_
        self.assertAllEqual(clustering_labels, expected)
